#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Provides class BaseVisualStim and mixins; subclass to get visual stimuli
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path
from statistics import mean
from psychopy.colors import Color, colorSpaces
from psychopy.layout import Vector, Position, Size, Vertices, unitTypes

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+

import pyglet
pyglet.options['debug_gl'] = False
GL = pyglet.gl

try:
    from PIL import Image
except ImportError:
    from . import Image

import copy
import sys
import os
from psychopy import logging

# tools must only be imported *after* event or MovieStim breaks on win32
# (JWP has no idea why!)
from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import (attributeSetter, logAttrib,
                                           setAttribute)
from psychopy.tools.monitorunittools import (cm2pix, deg2pix, pix2cm,
                                             pix2deg, convertToPix)
from psychopy.visual.helpers import (pointInPolygon, polygonsOverlap,
                                     setColor, findImageFile)
from psychopy.tools.typetools import float_uint8
from psychopy.tools.arraytools import makeRadialMatrix, createLumPattern
from psychopy.tools.colorspacetools import dkl2rgb, lms2rgb  # pylint: disable=W0611

from . import globalVars

import numpy
from numpy import pi

from psychopy.constants import NOT_STARTED, STARTED, STOPPED

reportNImageResizes = 5  # permitted number of resizes

"""
There are several base and mix-in visual classes for multiple inheritance:
  - MinimalStim:       non-visual house-keeping code common to all visual stim
        RatingScale inherits only from MinimalStim.
  - WindowMixin:       attributes/methods about the stim relative to
        a visual.Window.
  - LegacyVisualMixin: deprecated visual methods (eg, setRGB) added
        to BaseVisualStim
  - ColorMixin:        for Stim that need color methods (most, not Movie)
        color-related methods and attribs
  - ContainerMixin:    for stim that need polygon .contains() methods.
        Most need this, but not Text. .contains(), .overlaps()
  - TextureMixin:      for texture methods namely _createTexture
        (Grating, not Text)
        seems to work; caveat: There were issues in earlier (non-MI) versions
        of using _createTexture so it was pulled out of classes.
        Now it's inside classes again. Should be watched.
  - BaseVisualStim:    = Minimal + Window + Legacy. Furthermore adds c
        ommon attributes like orientation, opacity, contrast etc.

Typically subclass BaseVisualStim to create new visual stim classes, and add
mixin(s) as needed to add functionality.
"""


class MinimalStim:
    """Non-visual methods and attributes for BaseVisualStim and RatingScale.

    Includes: name, autoDraw, autoLog, status, __str__
    """

    def __init__(self, name=None, autoLog=None):
        if name not in (None, ''):
            self.__dict__['name'] = name
        else:
            self.__dict__['name'] = 'unnamed %s' % self.__class__.__name__
        self.status = NOT_STARTED
        self.autoLog = autoLog
        super(MinimalStim, self).__init__()
        if self.autoLog:
            msg = ("%s is calling MinimalStim.__init__() with autolog=True. "
                   "Set autoLog to True only at the end of __init__())")
            logging.warning(msg % self.__class__.__name__)

    def __str__(self, complete=False):
        """
        """
        if hasattr(self, '_initParams'):
            className = self.__class__.__name__
            paramStrings = []
            for param in self._initParams:
                if hasattr(self, param):
                    val = getattr(self, param)
                    valStr = repr(getattr(self, param))
                    if len(repr(valStr)) > 50 and not complete:
                        if val.__class__.__name__ == 'attributeSetter':
                            _name = val.__getattribute__.__class__.__name__
                        else:
                            _name = val.__class__.__name__
                        valStr = "%s(...)" % _name
                else:
                    valStr = 'UNKNOWN'
                paramStrings.append("%s=%s" % (param, valStr))
            # this could be used if all params are known to exist:
            # paramStrings = ["%s=%s" %(param, getattr(self, param))
            #     for param in self._initParams]
            params = ", ".join(paramStrings)
            s = "%s(%s)" % (className, params)
        else:
            s = object.__repr__(self)
        return s

    # Might seem simple at first, but this ensures that "name" attribute
    # appears in docs and that name setting and updating is logged.
    @attributeSetter
    def name(self, value):
        """The name (`str`) of the object to be using during logged messages
        about this stim. If you have multiple stimuli in your experiment this
        really helps to make sense of log files!

        If name = None your stimulus will be called "unnamed <type>", e.g.
        visual.TextStim(win) will be called "unnamed TextStim" in the logs.
        """
        self.__dict__['name'] = value

    @attributeSetter
    def autoDraw(self, value):
        """Determines whether the stimulus should be automatically drawn
        on every frame flip.

        Value should be: `True` or `False`. You do NOT need to set this
        on every frame flip!
        """
        self.__dict__['autoDraw'] = value
        toDraw = self.win._toDraw
        toDrawDepths = self.win._toDrawDepths
        beingDrawn = (self in toDraw)
        if value == beingDrawn:
            return  # nothing to do
        elif value:
            # work out where to insert the object in the autodraw list
            depthArray = numpy.array(toDrawDepths)
            # all indices where true:
            iis = numpy.where(depthArray < self.depth)[0]
            if len(iis):  # we featured somewhere before the end of the list
                toDraw.insert(iis[0], self)
                toDrawDepths.insert(iis[0], self.depth)
            else:
                toDraw.append(self)
                toDrawDepths.append(self.depth)
            # Add to editable list (if needed)
            self.win.addEditable(self)
            # Mark as started
            self.status = STARTED
        elif value == False:
            # remove from autodraw lists
            toDrawDepths.pop(toDraw.index(self))  # remove from depths
            toDraw.remove(self)  # remove from draw list
            # Remove from editable list (if needed)
            self.win.removeEditable(self)
            # Mark as stopped
            self.status = STOPPED

    def setAutoDraw(self, value, log=None):
        """Sets autoDraw. Usually you can use 'stim.attribute = value'
        syntax instead, but use this method to suppress the log message.
        """
        setAttribute(self, 'autoDraw', value, log)

    @attributeSetter
    def autoLog(self, value):
        """Whether every change in this stimulus should be auto logged.

        Value should be: `True` or `False`. Set to `False` if your stimulus is
        updating frequently (e.g. updating its position every frame) and you
        want to avoid swamping the log file with messages that aren't likely to
        be useful.
        """
        self.__dict__['autoLog'] = value

    def setAutoLog(self, value=True, log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message.
        """
        setAttribute(self, 'autoLog', value, log)


class LegacyVisualMixin:
    """Class to hold deprecated visual methods and attributes.

    Intended only for use as a mixin class for BaseVisualStim, to maintain
    backwards compatibility while reducing clutter in class BaseVisualStim.
    """
    # def __init__(self):
    #    super(LegacyVisualMixin, self).__init__()

    def _calcSizeRendered(self):
        """DEPRECATED in 1.80.00. This functionality is now handled
        by _updateVertices() and verticesPix
        """
        # raise DeprecationWarning, "_calcSizeRendered() was deprecated in
        # 1.80.00. This functionality is now handled by _updateVertices()
        # and verticesPix"
        if self.units in ['norm', 'pix', 'height']:
            self._sizeRendered = copy.copy(self.size)
        elif self.units in ['deg', 'degs']:
            self._sizeRendered = deg2pix(self.size, self.win.monitor)
        elif self.units == 'cm':
            self._sizeRendered = cm2pix(self.size, self.win.monitor)
        else:
            logging.error("Stimulus units should be 'height', 'norm', "
                          "'deg', 'cm' or 'pix', not '%s'" % self.units)

    def _calcPosRendered(self):
        """DEPRECATED in 1.80.00. This functionality is now handled
        by _updateVertices() and verticesPix.
        """
        # raise DeprecationWarning, "_calcSizeRendered() was deprecated
        #  in 1.80.00. This functionality is now handled by
        # _updateVertices() and verticesPix"
        if self.units in ['norm', 'pix', 'height']:
            self._posRendered = copy.copy(self.pos)
        elif self.units in ['deg', 'degs']:
            self._posRendered = deg2pix(self.pos, self.win.monitor)
        elif self.units == 'cm':
            self._posRendered = cm2pix(self.pos, self.win.monitor)

    def _getPolyAsRendered(self):
        """DEPRECATED. Return a list of vertices as rendered.
        """
        oriRadians = numpy.radians(self.ori)
        sinOri = numpy.sin(-oriRadians)
        cosOri = numpy.cos(-oriRadians)
        x = (self._verticesRendered[:, 0] * cosOri -
             self._verticesRendered[:, 1] * sinOri)
        y = (self._verticesRendered[:, 0] * sinOri +
             self._verticesRendered[:, 1] * cosOri)
        return numpy.column_stack((x, y)) + self._posRendered

    @attributeSetter
    def depth(self, value):
        """DEPRECATED, depth is now controlled simply by drawing order.
        """
        self.__dict__['depth'] = value


class LegacyForeColorMixin:
    """
    Mixin class to give an object all of the legacy functions for setting foreground color
    """
    def setDKL(self, color, operation=''):
        """DEPRECATED since v1.60.05: Please use the `color` attribute
        """
        self.setForeColor(color, 'dkl', operation)

    def setLMS(self, color, operation=''):
        """DEPRECATED since v1.60.05: Please use the `color` attribute
        """
        self.setForeColor(color, 'lms', operation)

    @property
    def foreRGB(self):
        """
        DEPRECATED: Legacy property for setting the foreground color of a stimulus in RGB, instead use `obj._foreColor.rgb`
        """
        return self._foreColor.rgb

    @foreRGB.setter
    def foreRGB(self, value):
        self.foreColor = Color(value, 'rgb')

    @property
    def RGB(self):
        """
        DEPRECATED: Legacy property for setting the foreground color of a stimulus in RGB, instead use `obj._foreColor.rgb`
        """
        return self.foreRGB

    @RGB.setter
    def RGB(self, value):
        self.foreRGB = value

    def setRGB(self, color, operation='', log=None):
        """
        DEPRECATED: Legacy setter for foreground RGB, instead set `obj._foreColor.rgb`
        """
        self.setForeColor(color, 'rgb', operation, log)

    def setForeRGB(self, color, operation='', log=None):
        """
        DEPRECATED: Legacy setter for foreground RGB, instead set `obj._foreColor.rgb`
        """
        self.setForeColor(color, 'rgb', operation, log)

    @property
    def foreColorSpace(self):
        """Deprecated, please use colorSpace to set color space for the entire
        object.
        """
        return self.colorSpace

    @foreColorSpace.setter
    def foreColorSpace(self, value):
        logging.warning(
            "Setting color space by attribute rather than by object is deprecated. Value of foreColorSpace has been assigned to colorSpace.")
        self.colorSpace = value


class LegacyFillColorMixin:
    """
    Mixin class to give an object all of the legacy functions for setting fill color
    """
    @property
    def fillRGB(self):
        """
        DEPRECATED: Legacy property for setting the fill color of a stimulus in RGB, instead use `obj._fillColor.rgb`
        """
        return self._fillColor.rgb

    @fillRGB.setter
    def fillRGB(self, value):
        self.fillColor = Color(value, 'rgb')

    @property
    def backRGB(self):
        """
        DEPRECATED: Legacy property for setting the fill color of a stimulus in RGB, instead use `obj._fillColor.rgb`
        """
        return self.fillRGB

    @backRGB.setter
    def backRGB(self, value):
        self.fillRGB = value

    def setFillRGB(self, color, operation='', log=None):
        """
        DEPRECATED: Legacy setter for fill RGB, instead set `obj._fillColor.rgb`
        """
        self.setFillColor(color, 'rgb', operation, log)

    def setBackRGB(self, color, operation='', log=None):
        """
        DEPRECATED: Legacy setter for fill RGB, instead set `obj._fillColor.rgb`
        """
        self.setFillColor(color, 'rgb', operation, log)

    @property
    def fillColorSpace(self):
        """Deprecated, please use colorSpace to set color space for the entire
        object.
        """
        return self.colorSpace

    @fillColorSpace.setter
    def fillColorSpace(self, value):
        logging.warning("Setting color space by attribute rather than by object is deprecated. Value of fillColorSpace has been assigned to colorSpace.")
        self.colorSpace = value

    @property
    def backColorSpace(self):
        """Deprecated, please use colorSpace to set color space for the entire
        object.
        """
        return self.colorSpace

    @backColorSpace.setter
    def backColorSpace(self, value):
        logging.warning(
            "Setting color space by attribute rather than by object is deprecated. Value of backColorSpace has been assigned to colorSpace.")
        self.colorSpace = value


class LegacyBorderColorMixin:
    """
    Mixin class to give an object all of the legacy functions for setting border color
    """
    @property
    def borderRGB(self):
        """
        DEPRECATED: Legacy property for setting the border color of a stimulus in RGB, instead use `obj._borderColor.rgb`
        """
        return self._borderColor.rgb

    @borderRGB.setter
    def borderRGB(self, value):
        self.borderColor = Color(value, 'rgb')

    @property
    def lineRGB(self):
        """
        DEPRECATED: Legacy property for setting the border color of a stimulus in RGB, instead use `obj._borderColor.rgb`
        """
        return self.borderRGB

    @lineRGB.setter
    def lineRGB(self, value):
        self.borderRGB = value

    def setBorderRGB(self, color, operation='', log=None):
        """
        DEPRECATED: Legacy setter for border RGB, instead set `obj._borderColor.rgb`
        """
        self.setBorderColor(color, 'rgb', operation, log)

    def setLineRGB(self, color, operation='', log=None):
        """
        DEPRECATED: Legacy setter for border RGB, instead set `obj._borderColor.rgb`
        """
        self.setBorderColor(color, 'rgb', operation, log)

    @property
    def borderColorSpace(self):
        """Deprecated, please use colorSpace to set color space for the entire
        object
        """
        return self.colorSpace

    @borderColorSpace.setter
    def borderColorSpace(self, value):
        logging.warning(
            "Setting color space by attribute rather than by object is deprecated. Value of borderColorSpace has been assigned to colorSpace.")
        self.colorSpace = value

    @property
    def lineColorSpace(self):
        """Deprecated, please use colorSpace to set color space for the entire
        object
        """
        return self.colorSpace

    @lineColorSpace.setter
    def lineColorSpace(self, value):
        logging.warning(
            "Setting color space by attribute rather than by object is deprecated. Value of lineColorSpace has been assigned to colorSpace.")
        self.colorSpace = value


class LegacyColorMixin(LegacyForeColorMixin, LegacyFillColorMixin, LegacyBorderColorMixin):
    """
    Mixin class to give an object all of the legacy functions for setting all colors (fore, fill and border
    """


class BaseColorMixin:
    """
    Mixin class giving base color methods (e.g. colorSpace) which are needed for any color stuff.
    """
    @property
    def colorSpace(self):
        """The name of the color space currently being used

        Value should be: a string or None

        For strings and hex values this is not needed.
        If None the default colorSpace for the stimulus is
        used (defined during initialisation).

        Please note that changing colorSpace does not change stimulus
        parameters. Thus you usually want to specify colorSpace before
        setting the color. Example::

            # A light green text
            stim = visual.TextStim(win, 'Color me!',
                                   color=(0, 1, 0), colorSpace='rgb')

            # An almost-black text
            stim.colorSpace = 'rgb255'

            # Make it light green again
            stim.color = (128, 255, 128)
        """
        if hasattr(self, '_colorSpace'):
            return self._colorSpace
        else:
            return 'rgba'

    @colorSpace.setter
    def colorSpace(self, value):
        if value in colorSpaces:
            self._colorSpace = value
        else:
            logging.error(f"'{value}' is not a valid color space")

    @property
    def contrast(self):
        """A value that is simply multiplied by the color.

        Value should be: a float between -1 (negative) and 1 (unchanged).
            :ref:`Operations <attrib-operations>` supported.

        Set the contrast of the stimulus, i.e. scales how far the stimulus
        deviates from the middle grey. You can also use the stimulus
        `opacity` to control contrast, but that cannot be negative.

        Examples::

            stim.contrast =  1.0  # unchanged contrast
            stim.contrast =  0.5  # decrease contrast
            stim.contrast =  0.0  # uniform, no contrast
            stim.contrast = -0.5  # slightly inverted
            stim.contrast = -1.0  # totally inverted

        Setting contrast outside range -1 to 1 is permitted, but may
        produce strange results if color values exceeds the monitor limits.::

            stim.contrast =  1.2  # increases contrast
            stim.contrast = -1.2  # inverts with increased contrast

        """
        if hasattr(self, '_foreColor'):
            return self._foreColor.contrast

    @contrast.setter
    def contrast(self, value):
        if hasattr(self, '_foreColor'):
            self._foreColor.contrast = value
        if hasattr(self, '_fillColor'):
            self._fillColor.contrast = value
        if hasattr(self, '_borderColor'):
            self._borderColor.contrast = value

    def setContrast(self, newContrast, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        if newContrast is not None:
            self.contrast = newContrast
        if operation in ['', '=']:
            self.contrast = newContrast
        elif operation in ['+']:
            self.contrast += newContrast
        elif operation in ['-']:
            self.contrast -= newContrast
        else:
            logging.error(f"Operation '{operation}' not recognised.")

    def _getDesiredRGB(self, rgb, colorSpace, contrast):
        """ Convert color to RGB while adding contrast.
        Requires self.rgb, self.colorSpace and self.contrast
        """
        col = Color(rgb, colorSpace)
        col.contrast *= contrast or 0
        return col.render('rgb')

    def updateColors(self):
        """Placeholder method to update colours when set externally, for example updating the `pallette` attribute of
        a textbox"""
        return


class ForeColorMixin(BaseColorMixin, LegacyForeColorMixin):
    """
    Mixin class for visual stim that need fore color.
    """
    @property
    def foreColor(self):
        """Foreground color of the stimulus

        Value should be one of:
            + string: to specify a :ref:`colorNames`. Any of the standard
              html/X11 `color names
              <http://www.w3schools.com/html/html_colornames.asp>`
              can be used.
            + :ref:`hexColors`
            + numerically: (scalar or triplet) for DKL, RGB or
                other :ref:`colorspaces`. For
                these, :ref:`operations <attrib-operations>` are supported.

        When color is specified using numbers, it is interpreted with
        respect to the stimulus' current colorSpace. If color is given as a
        single value (scalar) then this will be applied to all 3 channels.

        Examples
        --------
        For whatever stim you have::

            stim.color = 'white'
            stim.color = 'RoyalBlue'  # (the case is actually ignored)
            stim.color = '#DDA0DD'  # DDA0DD is hexadecimal for plum
            stim.color = [1.0, -1.0, -1.0]  # if stim.colorSpace='rgb':
                            # a red color in rgb space
            stim.color = [0.0, 45.0, 1.0]  # if stim.colorSpace='dkl':
                            # DKL space with elev=0, azimuth=45
            stim.color = [0, 0, 255]  # if stim.colorSpace='rgb255':
                            # a blue stimulus using rgb255 space
            stim.color = 255  # interpreted as (255, 255, 255)
                              # which is white in rgb255.


        :ref:`Operations <attrib-operations>` work as normal for all numeric
        colorSpaces (e.g. 'rgb', 'hsv' and 'rgb255') but not for strings, like
        named and hex. For example, assuming that colorSpace='rgb'::

            stim.color += [1, 1, 1]  # increment all guns by 1 value
            stim.color *= -1  # multiply the color by -1 (which in this
                                # space inverts the contrast)
            stim.color *= [0.5, 0, 1]  # decrease red, remove green, keep blue

        You can use `setColor` if you want to set color and colorSpace in one
        line. These two are equivalent::

            stim.setColor((0, 128, 255), 'rgb255')
            # ... is equivalent to
            stim.colorSpace = 'rgb255'
            stim.color = (0, 128, 255)
        """
        if hasattr(self, '_foreColor'):
            return self._foreColor.render(self.colorSpace)

    @foreColor.setter
    def foreColor(self, value):
        if isinstance(value, Color):
            # If supplied with a Color object, set as that
            self._foreColor = value
        else:
            # Otherwise, make a new Color object
            self._foreColor = Color(value, self.colorSpace, contrast=self.contrast)
        if not self._foreColor:
            self._foreColor = Color()
            logging.error(f"'{value}' is not a valid {self.colorSpace} color")

    @property
    def color(self):
        """Alternative way of setting `foreColor`."""
        return self.foreColor

    @color.setter
    def color(self, value):
        self.foreColor = value

    def setForeColor(self, color, colorSpace=None, operation='', log=None):
        """Hard setter for foreColor, allows suppression of the log message,
        simultaneous colorSpace setting and calls update methods.
        """
        setColor(obj=self, colorAttrib="foreColor", color=color, colorSpace=colorSpace or self.colorSpace, operation=operation)
        # Trigger color update for components like Textbox which have different behaviours for a hard setter
        self.updateColors()

    def setColor(self, color, colorSpace=None, operation='', log=None):
        self.setForeColor(color, colorSpace=colorSpace, operation=operation, log=log)


class FillColorMixin(BaseColorMixin, LegacyFillColorMixin):
    """
    Mixin class for visual stim that need fill color.
    """

    @property
    def fillColor(self):
        """Set the fill color for the shape."""
        if hasattr(self, '_fillColor'):
            return getattr(self._fillColor, self.colorSpace)  # return self._fillColor.render(self.colorSpace)

    @fillColor.setter
    def fillColor(self, value):
        if isinstance(value, Color):
            # If supplied with a color object, set as that
            self._fillColor = value
        else:
            # Otherwise, make a new Color object
            self._fillColor = Color(value, self.colorSpace, contrast=self.contrast)
        if not self._fillColor:
            # If given an invalid color, set as transparent and log error
            self._fillColor = Color()
            logging.error(f"'{value}' is not a valid {self.colorSpace} color")

    @property
    def backColor(self):
        """Alternative way of setting fillColor"""
        return self.fillColor

    @backColor.setter
    def backColor(self, value):
        self.fillColor = value

    def setFillColor(self, color, colorSpace=None, operation='', log=None):
        """Hard setter for fillColor, allows suppression of the log message,
        simultaneous colorSpace setting and calls update methods.
        """
        setColor(obj=self, colorAttrib="fillColor", color=color, colorSpace=colorSpace or self.colorSpace, operation=operation)
        # Trigger color update for components like Textbox which have different behaviours for a hard setter
        self.updateColors()

    def setBackColor(self, color, colorSpace=None, operation='', log=None):
        self.setFillColor(color, colorSpace=None, operation='', log=None)


class BorderColorMixin(BaseColorMixin, LegacyBorderColorMixin):
    @property
    def borderColor(self):
        if hasattr(self, '_borderColor'):
            return self._borderColor.render(self.colorSpace)

    @borderColor.setter
    def borderColor(self, value):
        if isinstance(value, Color):
            # If supplied with a color object, set as that
            self._borderColor = value
        else:
            # If supplied with a valid color, use it to make a color object
            self._borderColor = Color(value, self.colorSpace, contrast=self.contrast)
        if not self._borderColor:
            # If given an invalid color, set as transparent and log error
            self._borderColor = Color()
            logging.error(f"'{value}' is not a valid {self.colorSpace} color")

    @property
    def lineColor(self):
        """Alternative way of setting `borderColor`."""
        return self.borderColor

    @lineColor.setter
    def lineColor(self, value):
        self.borderColor = value

    def setBorderColor(self, color, colorSpace=None, operation='', log=None):
        """Hard setter for `fillColor`, allows suppression of the log message,
        simultaneous colorSpace setting and calls update methods.
        """
        setColor(obj=self, colorAttrib="borderColor", color=color, colorSpace=colorSpace or self.colorSpace, operation=operation)
        # Trigger color update for components like Textbox which have different behaviours for a hard setter
        self.updateColors()

    def setLineColor(self, color, colorSpace=None, operation='', log=None):
        self.setBorderColor(color, colorSpace=None, operation='', log=None)


class ColorMixin(ForeColorMixin, FillColorMixin, BorderColorMixin):
    """
    Mixin class for visual stim that need fill, fore and border color.
    """


class ContainerMixin:
    """Mixin class for visual stim that have verticesPix attrib
    and .contains() methods.
    """

    def __init__(self):
        super(ContainerMixin, self).__init__()
        self._verticesBase = numpy.array(
            [[0.5, -0.5], [-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5]])  # sqr
        self._borderBase = numpy.array(
            [[0.5, -0.5], [-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5]])  # sqr
        self._rotationMatrix = [[1., 0.], [0., 1.]]  # no rotation by default

    @property
    def verticesPix(self):
        """This determines the coordinates of the vertices for the
        current stimulus in pixels, accounting for size, ori, pos and units
        """
        # because this is a property getter we can check /on-access/ if it
        # needs updating :-)
        if self._needVertexUpdate:
            self._updateVertices()
        return self.__dict__['verticesPix']

    @property
    def _borderPix(self):
        """Allows for a dynamic border that differs from self.vertices, gets
        updated dynamically with identical transformations.
        """
        if not hasattr(self, 'border'):
            msg = "%s._borderPix requested without .border" % self.name
            logging.error(msg)
            raise AttributeError(msg)
        if self._needVertexUpdate:
            self._updateVertices()
        return self.__dict__['_borderPix']

    def _updateVertices(self):
        """Sets Stim.verticesPix and ._borderPix from pos, size, ori,
        flipVert, flipHoriz
        """
        # Get vertices from available attribs
        if hasattr(self, '_tesselVertices'):
            # Shapes need to render from this
            verts = self._tesselVertices
        elif hasattr(self, "_vertices"):
            # Non-shapes should use base vertices object
            verts = self._vertices
        else:
            # We'll settle for base verts array
            verts = self.vertices

        # Convert to a vertices object if not already
        if not isinstance(verts, Vertices):
            verts = Vertices(verts, obj=self)

        # If needed, sub in missing values for flip and anchor
        if hasattr(self, "flip"):
            verts.flip = self.flip
        if hasattr(self, "anchor"):
            verts.anchor = self.anchor
        # Supply current object's pos and size objects
        verts._size = self._size
        verts._pos = self._pos
        # Apply rotation
        verts = (verts.pix - self._pos.pix).dot(self._rotationMatrix) + self._pos.pix
        if hasattr(self, "_vertices"):
            borderVerts = (self._vertices.pix - self._pos.pix).dot(self._rotationMatrix) + self._pos.pix
        else:
            borderVerts = verts
        # Set values
        self.__dict__['verticesPix'] = verts
        self.__dict__['_borderPix'] = borderVerts
        # Mark as updated
        self._needVertexUpdate = False
        self._needUpdate = True  # but we presumably need to update the list

    def contains(self, x, y=None, units=None):
        """Returns True if a point x,y is inside the stimulus' border.

        Can accept variety of input options:
            + two separate args, x and y
            + one arg (list, tuple or array) containing two vals (x,y)
            + an object with a getPos() method that returns x,y, such
                as a :class:`~psychopy.event.Mouse`.

        Returns `True` if the point is within the area defined either by its
        `border` attribute (if one defined), or its `vertices` attribute if
        there is no .border. This method handles
        complex shapes, including concavities and self-crossings.

        Note that, if your stimulus uses a mask (such as a Gaussian) then
        this is not accounted for by the `contains` method; the extent of the
        stimulus is determined purely by the size, position (pos), and
        orientation (ori) settings (and by the vertices for shape stimuli).

        See Coder demos: shapeContains.py
        See Coder demos: shapeContains.py
        """
        # get the object in pixels
        if hasattr(x, 'border'):
            xy = x._borderPix  # access only once - this is a property
            units = 'pix'  # we can forget about the units
        elif hasattr(x, 'verticesPix'):
            # access only once - this is a property (slower to access)
            xy = x.verticesPix
            units = 'pix'  # we can forget about the units
        elif hasattr(x, 'getPos'):
            xy = x.getPos()
            units = x.units
        elif type(x) in [list, tuple, numpy.ndarray]:
            xy = numpy.array(x)
        else:
            xy = numpy.array((x, y))
        # try to work out what units x,y has
        if units is None:
            if hasattr(xy, 'units'):
                units = xy.units
            else:
                units = self.units
        if units != 'pix':
            xy = convertToPix(xy, pos=(0, 0), units=units, win=self.win)
        # ourself in pixels
        if hasattr(self, 'border'):
            poly = self._borderPix  # e.g., outline vertices
        elif hasattr(self, 'boundingBox'):
            if abs(self.ori) > 0.1:
                raise RuntimeError("TextStim.contains() doesn't currently "
                                   "support rotated text.")
            w, h = self.boundingBox  # e.g., outline vertices
            x, y = self.posPix
            poly = numpy.array([[x+w/2, y-h/2], [x-w/2, y-h/2],
                                [x-w/2, y+h/2], [x+w/2, y+h/2]])
        else:
            poly = self.verticesPix  # e.g., tessellated vertices

        return pointInPolygon(xy[0], xy[1], poly=poly)

    def overlaps(self, polygon):
        """Returns `True` if this stimulus intersects another one.

        If `polygon` is another stimulus instance, then the vertices
        and location of that stimulus will be used as the polygon.
        Overlap detection is typically very good, but it
        can fail with very pointy shapes in a crossed-swords configuration.

        Note that, if your stimulus uses a mask (such as a Gaussian blob)
        then this is not accounted for by the `overlaps` method; the extent
        of the stimulus is determined purely by the size, pos, and
        orientation settings (and by the vertices for shape stimuli).

        See coder demo, shapeContains.py
        """
        return polygonsOverlap(self, polygon)


class TextureMixin:
    """Mixin class for visual stim that have textures.

    Could move visual.helpers.setTexIfNoShaders() into here.

    Parameters
    ----------
    tex : Any
        Texture data. Value can be anything that resembles image data.
    id : int or :class:`~pyglet.gl.GLint`
        Texture ID.
    pixFormat : :class:`~pyglet.gl.GLenum` or int
        Pixel format to use, values can be `GL_ALPHA` or `GL_RGB`.
    stim : Any
        Stimulus object using the texture.
    res : int
        The resolution of the texture (unless a bitmap image is used).
    maskParams : dict or None
        Additional parameters to configure the mask used with this texture.
    forcePOW2 : bool
        Force the texture to be stored in a square memory area. For grating
        stimuli (anything that needs multiple cycles) `forcePOW2` should be
        set to be `True`. Otherwise the wrapping of the texture will not
        work.
    dataType : class:`~pyglet.gl.GLenum`, int or None
        None, `GL_UNSIGNED_BYTE`, `GL_FLOAT`. Only affects image files
        (numpy arrays will be float).
    wrapping : bool
        Enable wrapping of the texture. A texture will be set to repeat (or
        tile).

    """
    def _createTexture(self, tex, id, pixFormat, stim, res=128, maskParams=None,
                       forcePOW2=True, dataType=None, wrapping=True):

        # transform all variants of `None` to that, simplifies conditions below
        if tex in ["none", "None", "color"]:
            tex = None

        # Create an intensity texture, ranging -1:1.0
        notSqr = False  # most of the options will be creating a sqr texture
        wasImage = False  # change this if image loading works
        interpolate = stim.interpolate
        if dataType is None:
            if pixFormat == GL.GL_RGB:
                dataType = GL.GL_FLOAT
            else:
                dataType = GL.GL_UNSIGNED_BYTE

        # Fill out unspecified portions of maskParams with default values
        if maskParams is None:
            maskParams = {}
        # fringeWidth affects the proportion of the stimulus diameter that is
        # devoted to the raised cosine.
        allMaskParams = {'fringeWidth': 0.2, 'sd': 3}
        allMaskParams.update(maskParams)

        if type(tex) == numpy.ndarray:
            # handle a numpy array
            # for now this needs to be an NxN intensity array
            intensity = tex.astype(numpy.float32)
            if intensity.max() > 1 or intensity.min() < -1:
                logging.error('numpy arrays used as textures should be in '
                              'the range -1(black):1(white)')
            if len(tex.shape) == 3:
                wasLum = False
            else:
                wasLum = True
            # is it 1D?
            if tex.shape[0] == 1:
                stim._tex1D = True
                res = tex.shape[1]
            elif len(tex.shape) == 1 or tex.shape[1] == 1:
                stim._tex1D = True
                res = tex.shape[0]
            else:
                stim._tex1D = False
                # check if it's a square power of two
                maxDim = max(tex.shape)
                powerOf2 = 2 ** numpy.ceil(numpy.log2(maxDim))
                if (forcePOW2 and
                        (tex.shape[0] != powerOf2 or
                         tex.shape[1] != powerOf2)):
                    logging.error("Requiring a square power of two (e.g. "
                                  "16 x 16, 256 x 256) texture but didn't "
                                  "receive one")
                res = tex.shape[0]

            dataType = GL.GL_FLOAT
        elif tex in ("sin", "sqr", "saw", "tri", "sinXsin", "sqrXsqr", "circle",
                     "gauss", "cross", "radRamp", "raisedCos", None):
            if tex is None:
                res = 1
                wrapping = True  # override any wrapping setting for None

            # compute array of intensity value for desired pattern
            intensity = createLumPattern(tex, res, None, allMaskParams)
            wasLum = True
        else:
            if isinstance(tex, (str, Path)):
                # maybe tex is the name of a file:
                filename = findImageFile(tex, checkResources=True)
                if not filename:
                    msg = "Couldn't find image %s; check path? (tried: %s)"
                    logging.error(msg % (tex, os.path.abspath(tex)))
                    logging.flush()
                    raise IOError(msg % (tex, os.path.abspath(tex)))
                try:
                    im = Image.open(filename)
                    im = im.transpose(Image.FLIP_TOP_BOTTOM)
                except IOError:
                    msg = "Found file '%s', failed to load as an image"
                    logging.error(msg % (filename))
                    logging.flush()
                    msg = "Found file '%s' [= %s], failed to load as an image"
                    raise IOError(msg % (tex, os.path.abspath(tex)))
            else:
                # can't be a file; maybe its an image already in memory?
                try:
                    im = tex.copy().transpose(Image.FLIP_TOP_BOTTOM)
                except AttributeError:  # nope, not an image in memory
                    msg = "Couldn't make sense of requested image."
                    logging.error(msg)
                    logging.flush()
                    raise AttributeError(msg)
            # at this point we have a valid im
            stim._origSize = im.size
            wasImage = True
            # is it 1D?
            if im.size[0] == 1 or im.size[1] == 1:
                logging.error("Only 2D textures are supported at the moment")
            else:
                maxDim = max(im.size)
                powerOf2 = int(2**numpy.ceil(numpy.log2(maxDim)))
                if im.size[0] != powerOf2 or im.size[1] != powerOf2:
                    if not forcePOW2:
                        notSqr = True
                    elif globalVars.nImageResizes < reportNImageResizes:
                        msg = ("Image '%s' was not a square power-of-two ' "
                               "'image. Linearly interpolating to be %ix%i")
                        logging.warning(msg % (tex, powerOf2, powerOf2))
                        globalVars.nImageResizes += 1
                        im = im.resize([powerOf2, powerOf2], Image.BILINEAR)
                    elif globalVars.nImageResizes == reportNImageResizes:
                        logging.warning("Multiple images have needed resizing"
                                        " - I'll stop bothering you!")
                        im = im.resize([powerOf2, powerOf2], Image.BILINEAR)

            # is it Luminance or RGB?
            if pixFormat == GL.GL_ALPHA and im.mode != 'L':
                # we have RGB and need Lum
                wasLum = True
                im = im.convert("L")  # force to intensity (need if was rgb)
            elif im.mode == 'L':  # we have lum and no need to change
                wasLum = True
                dataType = GL.GL_FLOAT
            elif pixFormat == GL.GL_RGB:
                # we want RGB and might need to convert from CMYK or Lm
                # texture = im.tostring("raw", "RGB", 0, -1)
                im = im.convert("RGBA")
                wasLum = False
            else:
                raise ValueError('cannot determine if image is luminance or RGB')

            if dataType == GL.GL_FLOAT:
                # convert from ubyte to float
                # much faster to avoid division 2/255
                intensity = numpy.array(im).astype(
                    numpy.float32) * 0.0078431372549019607 - 1.0
            else:
                intensity = numpy.array(im)

        if pixFormat == GL.GL_RGB and wasLum and dataType == GL.GL_FLOAT:
            # grating stim on good machine
            # keep as float32 -1:1
            if (sys.platform != 'darwin' and
                    stim.win.glVendor.startswith('nvidia')):
                # nvidia under win/linux might not support 32bit float
                # could use GL_LUMINANCE32F_ARB here but check shader code?
                internalFormat = GL.GL_RGB16F_ARB
            else:
                # we've got a mac or an ATI card and can handle
                # 32bit float textures
                # could use GL_LUMINANCE32F_ARB here but check shader code?
                internalFormat = GL.GL_RGB32F_ARB
            # initialise data array as a float
            data = numpy.ones((intensity.shape[0], intensity.shape[1], 3),
                              numpy.float32)
            data[:, :, 0] = intensity  # R
            data[:, :, 1] = intensity  # G
            data[:, :, 2] = intensity  # B
        elif (pixFormat == GL.GL_RGB and
                wasLum and
                dataType != GL.GL_FLOAT):
            # was a lum image: stick with ubyte for speed
            internalFormat = GL.GL_RGB
            # initialise data array as a float
            data = numpy.ones((intensity.shape[0], intensity.shape[1], 3),
                              numpy.ubyte)
            data[:, :, 0] = intensity  # R
            data[:, :, 1] = intensity  # G
            data[:, :, 2] = intensity  # B
        elif pixFormat == GL.GL_RGB and dataType == GL.GL_FLOAT:
            # probably a custom rgb array or rgb image
            internalFormat = GL.GL_RGB32F_ARB
            data = intensity
        elif pixFormat == GL.GL_RGB:
            # not wasLum, not useShaders  - an RGB bitmap with no shader
            #  optionsintensity.min()
            internalFormat = GL.GL_RGB
            data = intensity  # float_uint8(intensity)
        elif pixFormat == GL.GL_ALPHA:
            internalFormat = GL.GL_ALPHA
            dataType = GL.GL_UNSIGNED_BYTE
            if wasImage:
                data = intensity
            else:
                data = float_uint8(intensity)
        else:
            raise ValueError("invalid or unsupported `pixFormat`")

        # check for RGBA textures
        if len(data.shape) > 2 and data.shape[2] == 4:
            if pixFormat == GL.GL_RGB:
                pixFormat = GL.GL_RGBA
            if internalFormat == GL.GL_RGB:
                internalFormat = GL.GL_RGBA
            elif internalFormat == GL.GL_RGB32F_ARB:
                internalFormat = GL.GL_RGBA32F_ARB
        texture = data.ctypes  # serialise

        # bind the texture in openGL
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, id)  # bind that name to the target
        # makes the texture map wrap (this is actually default anyway)
        if wrapping:
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT)
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT)
        else:
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP)
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP)
        # data from PIL/numpy is packed, but default for GL is 4 bytes
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        # important if using bits++ because GL_LINEAR
        # sometimes extrapolates to pixel vals outside range
        if interpolate:
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
            # GL_GENERATE_MIPMAP was only available from OpenGL 1.4
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_GENERATE_MIPMAP,
                               GL.GL_TRUE)
            GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, internalFormat,
                            data.shape[1], data.shape[0], 0,
                            pixFormat, dataType, texture)
        else:
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
            GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, internalFormat,
                            data.shape[1], data.shape[0], 0,
                            pixFormat, dataType, texture)

        GL.glTexEnvi(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE,
                     GL.GL_MODULATE)  # ?? do we need this - think not!
        # unbind our texture so that it doesn't affect other rendering
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        return wasLum

    def clearTextures(self):
        """Clear all textures associated with the stimulus.

        As of v1.61.00 this is called automatically during garbage collection
        of your stimulus, so doesn't need calling explicitly by the user.
        """
        GL.glDeleteTextures(1, self._texID)
        if hasattr(self, '_maskID'):
            GL.glDeleteTextures(1, self._maskID)

    @attributeSetter
    def mask(self, value):
        """The alpha mask (forming the shape of the image).

        This can be one of various options:
            * 'circle', 'gauss', 'raisedCos', 'cross'
            * **None** (resets to default)
            * the name of an image file (most formats supported)
            * a numpy array (1xN or NxN) ranging -1:1

        """
        self.__dict__['mask'] = value
        if self.__class__.__name__ == 'ImageStim':
            dataType = GL.GL_UNSIGNED_BYTE
        else:
            dataType = None
        self._createTexture(
            value, id=self._maskID, pixFormat=GL.GL_ALPHA, dataType=dataType,
            stim=self, res=self.texRes, maskParams=self.maskParams,
            wrapping=False)

    def setMask(self, value, log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message.
        """
        setAttribute(self, 'mask', value, log)

    @attributeSetter
    def texRes(self, value):
        """Power-of-two int. Sets the resolution of the mask and texture.
        texRes is overridden if an array or image is provided as mask.

        :ref:`Operations <attrib-operations>` supported.
        """
        self.__dict__['texRes'] = value

        # ... now rebuild textures (call attributeSetters without logging).
        if hasattr(self, 'tex'):
            setAttribute(self, 'tex', self.tex, log=False)
        if hasattr(self, 'mask'):
            setAttribute(self, 'mask', self.mask, log=False)

    @attributeSetter
    def maskParams(self, value):
        """Various types of input. Default to `None`.

        This is used to pass additional parameters to the mask if those are
        needed.

            - For 'gauss' mask, pass dict {'sd': 5} to control
                standard deviation.
            - For the 'raisedCos' mask, pass a dict: {'fringeWidth':0.2},
                where 'fringeWidth' is a parameter (float, 0-1), determining
                the proportion of the patch that will be blurred by the raised
                cosine edge."""
        self.__dict__['maskParams'] = value
        # call attributeSetter without log
        setAttribute(self, 'mask', self.mask, log=False)

    @attributeSetter
    def interpolate(self, value):
        """Whether to interpolate (linearly) the texture in the stimulus.

        If set to False then nearest neighbour will be used when needed,
        otherwise some form of interpolation will be used.
        """
        self.__dict__['interpolate'] = value


class WindowMixin:
    """Window-related attributes and methods.

    Used by BaseVisualStim, SimpleImageStim and ElementArrayStim.

    """
    @property
    def win(self):
        """The :class:`~psychopy.visual.Window` object in which the
        stimulus will be rendered by default. (required)

        Example, drawing same stimulus in two different windows and display
        simultaneously. Assuming that you have two windows and a stimulus
        (win1, win2 and stim)::

           stim.win = win1  # stimulus will be drawn in win1
           stim.draw()  # stimulus is now drawn to win1
           stim.win = win2  # stimulus will be drawn in win2
           stim.draw()  # it is now drawn in win2
           win1.flip(waitBlanking=False)  # do not wait for next
                        # monitor update
           win2.flip()  # wait for vertical blanking.

        Note that this just changes **default** window for stimulus.

        You could also specify window-to-draw-to when drawing::

           stim.draw(win1)
           stim.draw(win2)

        """
        return self.__dict__['win']

    @win.setter
    def win(self, value):
        self.__dict__['win'] = value
        # Update window ref in size and pos objects
        if hasattr(self, "_size") and isinstance(self._size, Vector):
            self._size.win = value
        if hasattr(self, "_pos") and isinstance(self._pos, Vector):
            self._pos.win = value

    @property
    def pos(self):
        if hasattr(self, "_pos"):
            return getattr(self._pos, self.units)

    @pos.setter
    def pos(self, value):
        self._pos = Position(value, units=self.units, win=self.win)

        if hasattr(self, "_vertices"):
            self._vertices._pos = self._pos

    @property
    def size(self):
        if hasattr(self, "_size"):
            return getattr(self._size, self.units)

    @size.setter
    def size(self, value):
        if value is None:
            value = (None, None)
        self._size = Size(value, units=self.units, win=self.win)

        if hasattr(self, "_vertices"):
            self._vertices._size = self._size

    @property
    def width(self):
        if len(self.size.shape) == 1:
            # Return first value if a 1d array
            return self.size[0]
        elif len(self.size.shape) == 2:
            # Return first column if a 2d array
            return self.size[:, 0]

    @width.setter
    def width(self, value):
        # Convert to a numpy array
        value = numpy.array(value)
        # Get original size
        size = self.size
        # Set size
        if len(self.size.shape) == 1:
            # Set first value if a 1d array
            size[0] = value
        elif len(self.size.shape) == 2:
            # Set first column if a 2d array
            size[:, 0] = value
        else:
            raise NotImplementedError(
                f"Cannot set height and width for {type(self).__name__} objects with more than 2 dimensions. "
                f"Please use .size instead."
            )

        self.size = size

    @property
    def height(self):
        if len(self.size.shape) == 1:
            # Return first value if a 1d array
            return self.size[1]
        elif len(self.size.shape) == 2:
            # Return first column if a 2d array
            return self.size[:, 1]

    @height.setter
    def height(self, value):
        # Convert to a numpy array
        value = numpy.array(value)
        # Get original size
        size = self.size
        # Set size
        if len(self.size.shape) == 1:
            # Set second value if a 1d array
            size[1] = value
        elif len(self.size.shape) == 2:
            # Set second column if a 2d array
            size[:, 1] = value
        else:
            raise NotImplementedError(
                f"Cannot set height and width for {type(self).__name__} objects with more than 2 dimensions. "
                f"Please use .size instead."
            )

        self.size = size

    @property
    def vertices(self):
        # Get or make Vertices object
        if hasattr(self, "_vertices"):
            verts = self._vertices
        else:
            # If not defined, assume vertices are just a square
            verts = self._vertices = Vertices(numpy.array([
                                [0.5, -0.5],
                                [-0.5, -0.5],
                                [-0.5, 0.5],
                                [0.5, 0.5],
                            ]), obj=self, flip=self.flip, anchor=self.anchor)
        return verts.base

    @vertices.setter
    def vertices(self, value):
        # If None, use defaut
        if value is None:
            value = [
                [0.5, -0.5],
                [-0.5, -0.5],
                [-0.5, 0.5],
                [0.5, 0.5],
            ]
        # Create Vertices object
        self._vertices = Vertices(value, obj=self, flip=self.flip, anchor=self.anchor)
        self._needVertexUpdate = True

    @property
    def flip(self):
        """
        1x2 array for flipping vertices along each axis; set as True to flip or False to not flip. If set as a single value, will duplicate across both axes. Accessing the protected attribute (`._flip`) will give an array of 1s and -1s with which to multiply vertices.
        """
        # Get base value
        if hasattr(self, "_flip"):
            flip = self._flip
        else:
            flip = numpy.array([[False, False]])
        # Convert from boolean
        return flip == -1

    @flip.setter
    def flip(self, value):
        if value is None:
            value = False
        # Convert to 1x2 numpy array
        value = numpy.array(value)
        value.resize((1, 2))
        # Ensure values were bool
        assert value.dtype == bool, "Flip values must be either a boolean (True/False) or an array of booleans"
        # Set as multipliers rather than bool
        self._flip = numpy.array([[
            -1 if value[0, 0] else 1,
            -1 if value[0, 1] else 1,
        ]])
        self._flipHoriz, self._flipVert = self._flip[0]
        # Apply to vertices
        if not hasattr(self, "_vertices"):
            self.vertices = None
        self._vertices.flip = self.flip
        # Mark as needing vertex update
        self._needVertexUpdate = True

    @property
    def flipHoriz(self):
        return self.flip[0][0]

    @flipHoriz.setter
    def flipHoriz(self, value):
        self.flip = [value, self.flip[0, 1]]

    @property
    def flipVert(self):
        return self.flip[0][1]

    @flipVert.setter
    def flipVert(self, value):
        self.flip = [self.flip[0, 0], value]

    @property
    def anchor(self):
        if hasattr(self, "_vertices"):
            return self._vertices.anchor
        elif hasattr(self, "_anchor"):
            # Return a backup value if there's no vertices yet
            return self._anchor

    @anchor.setter
    def anchor(self, value):
        if hasattr(self, "_vertices"):
            self._vertices.anchor = value
        else:
            # Set a backup value if there's no vertices yet
            self._anchor = value

    def setAnchor(self, value, log=None):
        setAttribute(self, 'anchor', value, log)

    @property
    def units(self):
        if hasattr(self, "_units"):
            return self._units
        else:
            return self.win.units

    @units.setter
    def units(self, value):
        """
        Units to use when drawing.

        Possible options are: None, 'norm', 'cm', 'deg', 'degFlat',
        'degFlatPos', or 'pix'.

        If None then the current units of the
        :class:`~psychopy.visual.Window` will be used.
        See :ref:`units` for explanation of other options.

        Note that when you change units, you don't change the stimulus
        parameters and it is likely to change appearance.

        Example::

            # This stimulus is 20% wide and 50% tall with respect to window
            stim = visual.PatchStim(win, units='norm', size=(0.2, 0.5)

            # This stimulus is 0.2 degrees wide and 0.5 degrees tall.
            stim.units = 'deg'

        """
        if value in unitTypes:
            self._units = value or self.win.units
            self._needVertexUpdate = True
        else:
            raise ValueError(f"Invalid unit type '{value}', must be one of: {unitTypes}")

    def draw(self):
        raise NotImplementedError('Stimulus classes must override '
                                  'visual.BaseVisualStim.draw')

    def _selectWindow(self, win):
        """Switch drawing to the specified window. Calls the window's
        _setCurrent() method which handles the switch.
        """
        win._setCurrent()

    def _updateList(self):
        """The user shouldn't need this method since it gets called
        after every call to .set()
        Chooses between using and not using shaders each call.
        """
        self._updateListShaders()


class BaseVisualStim(MinimalStim, WindowMixin, LegacyVisualMixin):
    """A template for a visual stimulus class.

    Actual visual stim like GratingStim, TextStim etc... are based on this.
    Not finished...?

    Methods defined here will override Minimal & Legacy, but best to avoid
    that for simplicity & clarity.

    """
    def __init__(self, win, units=None, name='', autoLog=None):
        self.autoLog = False  # just to start off during init, set at end
        self.win = win
        self.units = units
        self._rotationMatrix = [[1., 0.], [0., 1.]]  # no rotation by default
        # self.autoLog is set at end of MinimalStim.__init__
        super(BaseVisualStim, self).__init__(name=name, autoLog=autoLog)
        if self.autoLog:
            msg = ("%s is calling BaseVisualStim.__init__() with autolog=True"
                   ". Set autoLog to True only at the end of __init__())")
            logging.warning(msg % (self.__class__.__name__))

    @property
    def opacity(self):
        """Determines how visible the stimulus is relative to background.

        The value should be a single float ranging 1.0 (opaque) to 0.0
        (transparent). :ref:`Operations <attrib-operations>` are supported.
        Precisely how this is used depends on the :ref:`blendMode`.
        """
        alphas = []
        if hasattr(self, '_foreColor'):
            alphas.append(self._foreColor.alpha)
        if hasattr(self, '_fillColor'):
            alphas.append(self._fillColor.alpha)
        if hasattr(self, '_borderColor'):
            alphas.append(self._borderColor.alpha)
        if alphas:
            return mean(alphas)
        else:
            return 1

    @opacity.setter
    def opacity(self, value):
        # Setting opacity as a single value makes all colours the same opacity
        if value is None:
            # If opacity is set to be None, this indicates that each color should handle its own opacity
            return
        if hasattr(self, '_foreColor'):
            if self._foreColor != None:
                self._foreColor.alpha = value
        if hasattr(self, '_fillColor'):
            if self._fillColor != None:
                self._fillColor.alpha = value
        if hasattr(self, '_borderColor'):
            if self._borderColor != None:
                self._borderColor.alpha = value

    def updateOpacity(self):
        """Placeholder method to update colours when set externally, for example
        updating the `pallette` attribute of a textbox."""
        return

    @attributeSetter
    def ori(self, value):
        """The orientation of the stimulus (in degrees).

        Should be a single value (:ref:`scalar <attrib-scalar>`).
        :ref:`Operations <attrib-operations>` are supported.

        Orientation convention is like a clock: 0 is vertical, and positive
        values rotate clockwise. Beyond 360 and below zero values wrap
        appropriately.

        """
        self.__dict__['ori'] = float(value)
        radians = value * 0.017453292519943295
        sin, cos = numpy.sin, numpy.cos
        self._rotationMatrix = numpy.array([[cos(radians), -sin(radians)],
                                            [sin(radians), cos(radians)]])
        self._needVertexUpdate = True  # need to update update vertices
        self._needUpdate = True

    @property
    def size(self):
        """The size (width, height) of the stimulus in the stimulus
        :ref:`units <units>`

        Value should be :ref:`x,y-pair <attrib-xy>`,
        :ref:`scalar <attrib-scalar>` (applies to both dimensions)
        or None (resets to default). :ref:`Operations <attrib-operations>`
        are supported.

        Sizes can be negative (causing a mirror-image reversal) and can
        extend beyond the window.

        Example::

            stim.size = 0.8  # Set size to (xsize, ysize) = (0.8, 0.8)
            print(stim.size)  # Outputs array([0.8, 0.8])
            stim.size += (0.5, -0.5)  # make wider and flatter: (1.3, 0.3)

        Tip: if you can see the actual pixel range this corresponds to by
        looking at `stim._sizeRendered`
        """
        return WindowMixin.size.fget(self)

    @size.setter
    def size(self, value):
        # Supply default for None
        if value is None:
            value = Size((1, 1), units="height", win=self.win)
        # Duplicate single values
        if isinstance(value, (float, int)):
            value = (value, value)
        # Do setting
        WindowMixin.size.fset(self, value)
        # Mark any updates needed
        self._needVertexUpdate = True
        self._needUpdate = True
        if hasattr(self, '_calcCyclesPerStim'):
            self._calcCyclesPerStim()

    @property
    def pos(self):
        """
        The position of the center of the stimulus in the stimulus
        :ref:`units <units>`

        `value` should be an :ref:`x,y-pair <attrib-xy>`.
        :ref:`Operations <attrib-operations>` are also supported.

        Example::

            stim.pos = (0.5, 0)  # Set slightly to the right of center
            stim.pos += (0.5, -1)  # Increment pos rightwards and upwards.
                Is now (1.0, -1.0)
            stim.pos *= 0.2  # Move stim towards the center.
                Is now (0.2, -0.2)

        Tip: If you need the position of stim in pixels, you can obtain
        it like this::

            from psychopy.tools.monitorunittools import posToPix
            posPix = posToPix(stim)

        """
        return WindowMixin.pos.fget(self)

    @pos.setter
    def pos(self, value):
        # Supply defualt for None
        if value is None:
            value = Position((0, 0), units="height", win=self.win)
        # Do setting
        WindowMixin.pos.fset(self, value)
        # Mark any updates needed
        self._needVertexUpdate = True
        self._needUpdate = True

    def setPos(self, newPos, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message.
        """
        setAttribute(self, 'pos', val2array(newPos, False), log, operation)

    def setDepth(self, newDepth, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'depth', newDepth, log, operation)

    def setSize(self, newSize, operation='', units=None, log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        if units is None:
            # need to change this to create several units from one
            units = self.units
        setAttribute(self, 'size', val2array(newSize, False), log, operation)

    def setOri(self, newOri, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'ori', newOri, log, operation)

    def setOpacity(self, newOpacity, operation='', log=None):
        """Hard setter for opacity, allows the suppression of log messages and calls the update method
        """
        if operation in ['', '=']:
            self.opacity = newOpacity
        elif operation in ['+']:
            self.opacity += newOpacity
        elif operation in ['-']:
            self.opacity -= newOpacity
        else:
            logging.error(f"Operation '{operation}' not recognised.")
        # Trigger color update for components like Textbox which have different behaviours for a hard setter
        self.updateOpacity()

    def _set(self, attrib, val, op='', log=None):
        """DEPRECATED since 1.80.04 + 1.
        Use setAttribute() and val2array() instead.
        """
        # format the input value as float vectors
        if type(val) in [tuple, list, numpy.ndarray]:
            val = val2array(val)

        # Set attribute with operation and log
        setAttribute(self, attrib, val, log, op)

        # For DotStim
        if attrib in ('nDots', 'coherence'):
            self.coherence = round(self.coherence * self.nDots) / self.nDots



import pytest
from pathlib import Path

from psychopy import experiment


@pytest.mark.stdroutines
class TestStandaloneRoutines:
    @classmethod
    def setup_class(cls):
        cls.routines = experiment.getAllStandaloneRoutines()
        # Make basic experiments with one of each standalone routine
        cls.expPy = experiment.Experiment()
        cls.expJS = experiment.Experiment()

    def setup(self):
        """This setup is done for each test individually
        """
        pass

    def teardown(self):
        pass

    def test_writing(self):
        # Test both python and JS
        for target, exp in {"PsychoPy": self.expPy, "PsychoJS": self.expJS}.items():
            for name, routine in self.routines.items():
                rt = routine(exp)
                exp.addStandaloneRoutine(name, rt)
                exp.flow.addRoutine(rt, 0)
            # Compile
            exp.writeScript(target=target)
            # Remove routines
            for name, routine in exp.routines.copy().items():
                exp.flow.removeComponent(routine)
                del exp.routines[name]



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import socket
import os
import numbers  # numbers.Integral is like (int, long) but supports Py3
from psychopy import colors
from psychopy.tools import arraytools
from ..util import yload, yLoader, module_directory
from ..errors import print2err

# Takes a device configuration yaml dict and processes it based on the devices
# support_settings_values.yaml (which must be in the same directory as the
# Device class) to ensure all entries for the device setting are valid values.

class ValidationError(Exception):
    """Base class for exceptions in this module."""
    pass


class BooleanValueError(ValidationError):
    """Exception raised for errors when a bool was expected for the settings
    parameter value.

    Attributes:
        device_config_setting_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        msg -- explanation of the error

    """

    def __init__(self, device_param_name, value_given):
        self.msg = 'A bool value is required for the given Device configuration parameter'
        self.device_config_param_name = device_param_name
        self.value_given = value_given

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class StringValueError(ValidationError):
    """Exception raised for errors when a str was expected for the settings
    parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        device_config_param_constraints  -- the set of constraints that apply to the parameter.
        msg -- explanation of the error
    """

    def __init__(
            self,
            device_config_param_name,
            value_given,
            device_config_param_constraints):
        self.msg = 'A str value is required for the given Device configuration parameter that meets the specified constraints'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given
        self.device_config_param_constraints = device_config_param_constraints

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class FloatValueError(ValidationError):
    """Exception raised for errors when a float was expected for the settings
    parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        device_config_param_constraints  -- the set of constraints that apply to the parameter.
        msg -- explanation of the error

    """

    def __init__(
            self,
            device_config_param_name,
            value_given,
            device_config_param_constraints):
        self.msg = 'A float value is required for the given Device configuration parameter that meets the specified constraints'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given
        self.device_config_param_constraints = device_config_param_constraints

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class IntValueError(ValidationError):
    """Exception raised for errors when an int was expected for the settings
    parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        device_config_param_constraints  -- the set of constraints that apply to the parameter.
        msg -- explanation of the error

    """

    def __init__(
            self,
            device_config_param_name,
            value_given,
            device_config_param_constraints):
        self.msg = 'An int value is required for the given Device configuration parameter that meets the specified constraints'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given
        self.device_config_param_constraints = device_config_param_constraints

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class NumberValueError(ValidationError):
    """Exception raised for errors when an int OR float was expected for the
    settings parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        device_config_param_constraints  -- the set of constraints that apply to the parameter.
        msg -- explanation of the error

    """

    def __init__(
            self,
            device_config_param_name,
            value_given,
            device_config_param_constraints):
        self.msg = 'An int or float value is required for the given Device configuration parameter that meets the specified constraints'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given
        self.device_config_param_constraints = device_config_param_constraints

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class IpValueError(ValidationError):
    """Exception raised for errors when an IP address was expected for the
    settings parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        msg -- explanation of the error

    """

    def __init__(self, device_config_param_name, value_given):
        self.msg = 'An IP address value is required for the given Device configuration parameter.'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class ColorValueError(ValidationError):
    """Exception raised for errors when a color was expected for the settings
    parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        msg -- explanation of the error

    """

    def __init__(self, device_config_param_name, value_given):
        self.msg = 'A color value is required for the given Device configuration parameter.'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class DateStringValueError(ValidationError):
    """Exception raised for errors when a date string was expected for the
    settings parameter value.

    Attributes:
        device_config_param_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        msg -- explanation of the error

    """

    def __init__(self, device_config_param_name, value_given):
        self.msg = 'A date string value is required for the given Device configuration parameter.'
        self.device_config_param_name = device_config_param_name
        self.value_given = value_given

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n'.format(
            self.__class__.__name__, self.msg, self.device_config_param_name, self.value_given)


class NonSupportedValueError(ValidationError):
    """Exception raised when the configuration value provided does not match
    one of the possible valid Device configuration parameter values.

    Attributes:
        device_config_setting_name -- The name of the Device configuration parameter that has the error.
        value_given  -- the value read from the experiment configuration file.
        valid_values  -- the valid options for the configuration setting.
        msg -- explanation of the error

    """

    def __init__(self, device_param_name, value_given, valid_values):
        self.msg = 'A the provided value is not supported for the given Device configuration parameter'
        self.device_config_param_name = device_param_name
        self.value_given = value_given
        self.valid_values = valid_values

    def __str__(self):
        return '\n{0}:\n\tmsg: {1}\n\tparam_name: {2}\n\tvalue: {3}\n\tconstraints: {4}'.format(
            self.__class__.__name__,
            self.msg,
            self.device_config_param_name,
            self.value_given,
            self.valid_values)

MIN_VALID_STR_LENGTH = 1
MAX_VALID_STR_LENGTH = 1024

MIN_VALID_FLOAT_VALUE = 0.0
MAX_VALID_FLOAT_VALUE = 1000000.0

MIN_VALID_INT_VALUE = 0
MAX_VALID_INT_VALUE = 1000000

def is_sequence(arg):
    return hasattr(arg, "__getitem__") or hasattr(arg, "__iter__")

def isValidColor(config_param_name, color, constraints):
    """
    Return color if it is a valid psychopy color (regardless of color space)
    , otherwise raise error. Color value can be in hex, name, rgb, rgb255 format.
    """
    if isinstance(color, str):
        if color[0] == '#' or color[0:2].lower() == '0x':
            rgb255color = colors.hex2rgb255(color)
            if rgb255color is not None:
                return color
            else:
                raise ColorValueError(config_param_name, color)

        if color.lower() in colors.colorNames.keys():
            return color
        else:
            raise ColorValueError(config_param_name, color)
    if isinstance(color, (float, int)) or (is_sequence(color) and len(color) == 3):
        colorarray = arraytools.val2array(color, length=3)
        if colorarray is not None:
            return color
        else:
            raise ColorValueError(config_param_name, color)
    raise ColorValueError(config_param_name, color)

def isValidString(config_param_name, value, constraints):
    if isinstance(value, str):
        if value == constraints:
            # static string
            return value
        constraints.setdefault('min_length', MIN_VALID_STR_LENGTH)
        constraints.setdefault('max_length', MAX_VALID_STR_LENGTH)
        constraints.setdefault('first_char_alpha', False)
        min_length = int(constraints.get('min_length'))
        max_length = int(constraints.get('max_length'))
        first_char_alpha = bool(constraints.get('first_char_alpha'))

        if len(value) >= min_length:
            if len(value) <= max_length:
                if first_char_alpha is True and value[0].isalpha() is False:
                    raise StringValueError(
                        config_param_name, value, constraints)
                else:
                    return value

    elif int(constraints.get('min_length')) == 0 and value is None:
        return value

    raise StringValueError(config_param_name, value, constraints)


def isValidFloat(config_param_name, value, constraints):
    if isinstance(value, float):
        constraints.setdefault('min', MIN_VALID_FLOAT_VALUE)
        constraints.setdefault('max', MAX_VALID_FLOAT_VALUE)
        minv = float(constraints.get('min'))
        maxv = float(constraints.get('max'))

        if value >= minv:
            if value <= maxv:
                return value

    raise FloatValueError(config_param_name, value, constraints)


def isValidInt(config_param_name, value, constraints):
    if isinstance(value, numbers.Integral):
        constraints.setdefault('min', MIN_VALID_INT_VALUE)
        constraints.setdefault('max', MAX_VALID_INT_VALUE)
        minv = int(constraints.get('min'))
        maxv = int(constraints.get('max'))

        if value >= minv:
            if value <= maxv:
                return value

    raise IntValueError(config_param_name, value, constraints)

def isBool(config_param_name, value, valid_value):
    try:
        value = bool(value)
        return value
    except Exception:
        raise BooleanValueError(config_param_name, value)


def isValidIpAddress(config_param_name, value, valid_value):
    try:
        socket.inet_aton(value)
        return value
    except Exception:
        raise IpValueError(config_param_name, value)


def isValidList(config_param_name, value, constraints):
    try:
        min_length = constraints.get('min_length', 1)
        max_length = constraints.get('max_length', 128)

        if min_length == 0 and value is None or value == 'None':
            return value

        valid_values = constraints.get('valid_values', [])

        if len(valid_values) == 0:
            return value

        if not isinstance(value, (list, tuple)):
            if value not in valid_values:
                raise NonSupportedValueError(
                    config_param_name, value, constraints)
            elif min_length in [0, 1]:
                return value

        current_length = len(value)

        if current_length < min_length or current_length > max_length:
            raise NonSupportedValueError(config_param_name, value, constraints)

        for v in value:
            if v not in valid_values:
                raise NonSupportedValueError(config_param_name, v, valid_values)

        return value

    except Exception:
        raise NonSupportedValueError(config_param_name, value, constraints)


def isValueValid(config_param_name, value, valid_values):
    if isinstance(value, (list, tuple)):
        for v in value:
            if v not in valid_values:
                raise NonSupportedValueError(config_param_name, value, valid_values)
    elif value not in valid_values:
        raise NonSupportedValueError(config_param_name, value, valid_values)
    return value

CONFIG_VALIDATION_KEY_WORD_MAPPINGS = dict(
    IOHUB_STRING=isValidString,
    IOHUB_BOOL=isBool,
    IOHUB_FLOAT=isValidFloat,
    IOHUB_INT=isValidInt,
    IOHUB_LIST=isValidList,
    IOHUB_COLOR=isValidColor,
    IOHUB_IP_ADDRESS_V4=isValidIpAddress)
###############################################

_current_dir = module_directory(isValidString)


def buildConfigParamValidatorMapping(
        device_setting_validation_dict,
        param_validation_func_mapping,
        parent_name):
    for param_name, param_config in device_setting_validation_dict.items():
        current_param_path = None
        if parent_name is None:
            current_param_path = param_name
        else:
            current_param_path = '%s.%s' % (parent_name, param_name)

        keyword_validator_function = None
        if isinstance(param_name, str):
            keyword_validator_function = CONFIG_VALIDATION_KEY_WORD_MAPPINGS.get(
                param_name, None)

        if keyword_validator_function:
            param_validation_func_mapping[
                parent_name] = keyword_validator_function, param_config
            #print2err('ADDED MAPPING1: ', current_param_path, " ", isValueValid, " ", param_config)
        elif isinstance(param_config, str):
            keyword_validator_function = CONFIG_VALIDATION_KEY_WORD_MAPPINGS.get(
                param_config, None)
            if keyword_validator_function:
                param_validation_func_mapping[
                    current_param_path] = keyword_validator_function, {}
                #print2err('ADDED MAPPING2: ', current_param_path, " ", isValueValid, " ", param_config)
            else:
                param_validation_func_mapping[
                    current_param_path] = isValueValid, [param_config, ]
                #print2err('ADDED MAPPING3: ', current_param_path, " ", isValueValid, " ", param_config)
        elif isinstance(param_config, dict):
            buildConfigParamValidatorMapping(
                param_config,
                param_validation_func_mapping,
                current_param_path)
        elif isinstance(param_config, (list, tuple)):
            param_validation_func_mapping[
                current_param_path] = isValueValid, param_config
            #print2err('ADDED MAPPING4: ', current_param_path, " ", isValueValid, " ", param_config)
        else:
            param_validation_func_mapping[
                current_param_path] = isValueValid, [param_config, ]
            #print2err('ADDED MAPPING5: ', current_param_path, " ", isValueValid, " ", param_config)


def validateConfigDictToFuncMapping(
        param_validation_func_mapping,
        current_device_config,
        parent_param_path):
    validation_results = dict(errors=[], not_found=[])
    for config_param, config_value in current_device_config.items():
        if parent_param_path is None:
            current_param_path = config_param
        else:
            current_param_path = '%s.%s' % (parent_param_path, config_param)

        param_validation = param_validation_func_mapping.get(
            current_param_path, None)
        if param_validation:
            param_validation_func, constraints = param_validation
            try:
                param_value = param_validation_func(
                    current_param_path, config_value, constraints)
                current_device_config[config_param] = param_value
#                print2err("PARAM {0}, VALUE {1} is VALID.".format(current_param_path,param_value))
            except ValidationError:
                validation_results['errors'].append(
                    (config_param, config_value))
                #print2err("Device Config Validation Error: param: {0}, value: {1}\nError: {2}".format(config_param,config_value,e))

        elif isinstance(config_value, dict):
            validateConfigDictToFuncMapping(
                param_validation_func_mapping,
                config_value,
                current_param_path)
        else:
            validation_results['not_found'].append(
                (config_param, config_value))
    return validation_results


def validateDeviceConfiguration(
        relative_module_path,
        device_class_name,
        current_device_config):
    validation_file_path = os.path.join(_current_dir,
                                        relative_module_path[len('psychopy.iohub.devices.'):].replace('.', os.path.sep),
                                        'supported_config_settings_{0}.yaml'.format(device_class_name.lower()))
    if not os.path.exists(validation_file_path):
        validation_file_path = os.path.join(_current_dir,
                                            relative_module_path[len('psychopy.iohub.devices.'):].replace('.',
                                                                                                          os.path.sep),
                                            'supported_config_settings.yaml')
    device_settings_validation_dict = yload(open(validation_file_path, 'r'), Loader=yLoader)
    device_settings_validation_dict = device_settings_validation_dict[list(device_settings_validation_dict.keys())[0]]

    param_validation_func_mapping = dict()
    parent_config_param_path = None
    buildConfigParamValidatorMapping(device_settings_validation_dict, param_validation_func_mapping,
                                     parent_config_param_path)

    validation_results = validateConfigDictToFuncMapping(param_validation_func_mapping, current_device_config, None)

    return validation_results



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import copy
import psychopy
from .text import TextStim
from .rect import Rect
from psychopy.data.utils import importConditions, listFromString
from psychopy.visual.basevisual import (BaseVisualStim,
                                        ContainerMixin,
                                        ColorMixin)
from psychopy import logging, layout
from random import shuffle
from pathlib import Path

__author__ = 'Jon Peirce, David Bridges, Anthony Haffey'

from ..colors import Color

_REQUIRED = -12349872349873  # an unlikely int

# a dict of known fields with their default vals
_knownFields = {
    'index': None,  # optional field to index into the rows
    'itemText': _REQUIRED,  # (question used until 2020.2)
    'itemColor': None,
    'itemWidth': 1,  # fraction of the form
    'type': _REQUIRED,  # type of response box (see below)
    'options': ('Yes', 'No'),  # for choice box
    'ticks': None,#(1, 2, 3, 4, 5, 6, 7),
    'tickLabels': None,
    'font': None,
    # for rating/slider
    'responseWidth': 1,  # fraction of the form
    'responseColor': None,
    'markerColor': None,
    'layout': 'horiz',  # can be vert or horiz
}
_doNotSave = [
    'itemCtrl', 'responseCtrl',  # these genuinely can't be save
    'itemColor', 'itemWidth', 'options', 'ticks', 'tickLabels',  # not useful?
    'responseWidth', 'responseColor', 'layout',
]
_knownRespTypes = {
    'heading', 'description',  # no responses
    'rating', 'slider',  # slider is continuous
    'free text',
    'choice', 'radio'  # synonyms (radio was used until v2020.2)
}
_synonyms = {
    'itemText': 'questionText',
    'choice': 'radio',
    'free text': 'textBox'
}

# Setting debug to True will set the sub-elements on Form to be outlined in red, making it easier to determine their position
debug = False


class Form(BaseVisualStim, ContainerMixin, ColorMixin):
    """A class to add Forms to a `psychopy.visual.Window`

    The Form allows Psychopy to be used as a questionnaire tool, where
    participants can be presented with a series of questions requiring responses.
    Form items, defined as questions and response pairs, are presented
    simultaneously onscreen with a scrollable viewing window.

    Example
    -------
    survey = Form(win, items=[{}], size=(1.0, 0.7), pos=(0.0, 0.0))

    Parameters
    ----------
    win : psychopy.visual.Window
        The window object to present the form.
    items : List of dicts or csv or xlsx file
        a list of dicts or csv file should have the following key, value pairs / column headers:
                 "index": The item index as a number
                 "itemText": item question string,
                 "itemWidth": fraction of the form width 0:1
                 "type": type of rating e.g., 'radio', 'rating', 'slider'
                 "responseWidth": fraction of the form width 0:1,
                 "options": list of tick labels for options,
                 "layout": Response object layout e.g., 'horiz' or 'vert'
    textHeight : float
        Text height.
    size : tuple, list
        Size of form on screen.
    pos : tuple, list
        Position of form on screen.
    itemPadding : float
        Space or padding between form items.
    units : str
        units for stimuli - Currently, Form class only operates with 'height' units.
    randomize : bool
        Randomize order of Form elements
    """

    knownStyles = {
        'light': {
            'fillColor': [0.89, 0.89, 0.89],
            'borderColor': None,
            'itemColor': 'black',
            'responseColor': 'black',
            'markerColor': [0.89, -0.35, -0.28],
            'font': "Open Sans",
        },
        'dark': {
            'fillColor': [-0.19, -0.19, -0.14],
            'borderColor': None,
            'itemColor': 'white',
            'responseColor': 'white',
            'markerColor': [0.89, -0.35, -0.28],
            'font': "Open Sans",
        },
    }

    def __init__(self,
                 win,
                 name='default',
                 colorSpace='rgb',
                 fillColor=None,
                 borderColor=None,
                 itemColor='white',
                 responseColor='white',
                 markerColor='red',
                 items=None,
                 font=None,
                 textHeight=.02,
                 size=(.5, .5),
                 pos=(0, 0),
                 style=None,
                 itemPadding=0.05,
                 units='height',
                 randomize=False,
                 autoLog=True,
                 # legacy
                 color=None,
                 foreColor=None
                 ):

        super(Form, self).__init__(win, units, autoLog=False)
        self.win = win
        self.autoLog = autoLog
        self.name = name
        self.randomize = randomize
        self.items = self.importItems(items)
        self.size = size
        self._pos = pos
        self.itemPadding = itemPadding
        self.scrollSpeed = self.setScrollSpeed(self.items, 4)
        self.units = units
        self.depth = 0

        # Appearance
        self.colorSpace = colorSpace
        self.fillColor = fillColor
        self.borderColor = borderColor
        self.itemColor = itemColor
        self.responseColor = responseColor
        self.markerColor = markerColor
        if color:
            self.foreColor = color
        if foreColor:
            self.foreColor = color

        self.font = font or "Open Sans"

        self.textHeight = textHeight
        self._baseYpositions = []
        self.leftEdge = None
        self.rightEdge = None
        self.topEdge = None
        self._currentVirtualY = 0  # Y position in the virtual sheet
        self._vheight = 0  # Height of the virtual sheet
        self._decorations = []
        self._externalDecorations = []
        # Check units - only works with height units for now
        if self.win.units != 'height':
            logging.warning(
                "Form currently only formats correctly using height units. "
                "Please change the units in Experiment Settings to 'height'")

        self._complete = False

        # Create layout of form
        self._createItemCtrls()

        self.style = style

        if self.autoLog:
            logging.exp("Created {} = {}".format(self.name, repr(self)))

    def __repr__(self, complete=False):
        return self.__str__(complete=complete)  # from MinimalStim

    def importItems(self, items):
        """Import items from csv or excel sheet and convert to list of dicts.
        Will also accept a list of dicts.

        Note, for csv and excel files, 'options' must contain comma separated values,
        e.g., one, two, three. No parenthesis, or quotation marks required.

        Parameters
        ----------
        items :  Excel or CSV file, list of dicts
            Items used to populate the Form

        Returns
        -------
        List of dicts
            A list of dicts, where each list entry is a dict containing all fields for a single Form item
        """
        def _checkSynonyms(items, fieldNames):
            """Checks for updated names for fields (i.e. synonyms)"""

            replacedFields = set()
            for field in _synonyms:
                synonym = _synonyms[field]
                for item in items:
                    if synonym in item:
                        # convert to new name
                        item[field] = item[synonym]
                        del item[synonym]
                        replacedFields.add(field)
            for field in replacedFields:
                fieldNames.append(field)
                fieldNames.remove(_synonyms[field])
                logging.warning("Form {} included field no longer used {}. "
                                "Replacing with new name '{}'"
                                .format(self.name, _synonyms[field], field))

        def _checkRequiredFields(fieldNames):
            """Checks for required headings (do this after checking synonyms)"""
            for hdr in _knownFields:
                # is it required and/or present?
                if _knownFields[hdr] == _REQUIRED and hdr not in fieldNames:
                    raise ValueError("Missing header ({}) in Form ({}). "
                                     "Headers found were: {}"
                                     .format(hdr, self.name, fieldNames))


        def _checkTypes(types, itemText):
            """A nested function for testing the number of options given

            Raises ValueError if n Options not > 1
            """
            itemDiff = set([types]) - set(_knownRespTypes)

            for incorrItemType in itemDiff:
                if incorrItemType == _REQUIRED:
                    if self._itemsFile:
                        itemsFileStr =  ("in items file '{}'"
                                         .format(self._itemsFile))
                    else:
                        itemsFileStr = ""
                    msg = ("Item {}{} is missing a required "
                           "value for its response type. Permitted types are "
                           "{}.".format(itemText, itemsFileStr,
                                        _knownRespTypes))
                if self.autoLog:
                    logging.error(msg)
                raise ValueError(msg)

        def _addDefaultItems(items):
            """
            Adds default items when missing. Works in-place.

            Parameters
            ----------
            items : List of dicts
            headers : List of column headers for each item
            """
            def isPresent(d, field):
                # check if the field is there and not empty on this row
                return (field in d and d[field] not in [None, ''])

            missingHeaders = []
            defaultValues = _knownFields
            for index, item in enumerate(items):
                defaultValues['index'] = index
                for header in defaultValues:
                    # if header is missing of val is None or ''
                    if not isPresent(item, header):
                        oldHeader = header.replace('item', 'question')
                        if isPresent(item, oldHeader):
                            item[header] = item[oldHeader]
                            logging.warning(
                                "{} is a deprecated heading for Forms. "
                                "Use {} instead"
                                .format(oldHeader, header)
                            )
                            continue
                        # Default to colour scheme if specified
                        if defaultValues[header] in ['fg', 'bg', 'em']:
                            item[header] = self.color
                        else:
                            item[header] = defaultValues[header]
                        missingHeaders.append(header)

            msg = "Using default values for the following headers: {}".format(
                missingHeaders)
            if self.autoLog:
                logging.info(msg)

        if self.autoLog:
            logging.info("Importing items...")

        if not isinstance(items, list):
            # items is a conditions file
            self._itemsFile = Path(items)
            items, fieldNames = importConditions(items, returnFieldNames=True)
        else:  # we already have a list so lets find the fieldnames
            fieldNames = set()
            for item in items:
                fieldNames = fieldNames.union(item)
            fieldNames = list(fieldNames)  # convert to list at the end
            self._itemsFile = None

        _checkSynonyms(items, fieldNames)
        _checkRequiredFields(fieldNames)
        # Add default values if entries missing
        _addDefaultItems(items)

        # Convert options to list of strings
        for idx, item in enumerate(items):
            if item['ticks']:
                item['ticks'] = listFromString(item['ticks'])
            if 'tickLabels' in item and item['tickLabels']:
                item['tickLabels'] = listFromString(item['tickLabels'])
            if 'options' in item and item['options']:
                item['options'] = listFromString(item['options'])

        # Check types
        [_checkTypes(item['type'], item['itemText']) for item in items]
        # Check N options > 1
        # Randomise items if requested
        if self.randomize:
            shuffle(items)
        return items

    def setScrollSpeed(self, items, multiplier=2):
        """Set scroll speed of Form. Higher multiplier gives smoother, but
        slower scroll.

        Parameters
        ----------
        items : list of dicts
            Items used to populate the form
        multiplier : int (default=2)
            Number used to calculate scroll speed

        Returns
        -------
        int
            Scroll speed, calculated using N items by multiplier
        """
        return len(items) * multiplier

    def _getItemRenderedWidth(self, size):
        """Returns text width for item text based on itemWidth and Form width.

        Parameters
        ----------
        size : float, int
            The question width

        Returns
        -------
        float
            Wrap width for question text
        """
        return size * self.size[0] - (self.itemPadding * 2)

    def _setQuestion(self, item):
        """Creates TextStim object containing question

        Parameters
        ----------
        item : dict
            The dict entry for a single item

        Returns
        -------
        psychopy.visual.text.TextStim
            The textstim object with the question string
        questionHeight
            The height of the question bounding box as type float
        questionWidth
            The width of the question bounding box as type float
        """
        if self.autoLog:
            logging.exp(
                    u"Question text: {}".format(item['itemText']))

        if item['type'] == 'heading':
            letterScale = 1.5
            bold = True
        else:
            letterScale = 1.0
            bold = False
        w = self._getItemRenderedWidth(item['itemWidth'])
        question = psychopy.visual.TextBox2(
                self.win,
                text=item['itemText'],
                units=self.units,
                letterHeight=self.textHeight * letterScale,
                anchor='top-left',
                alignment='center-left',
                pos=(self.leftEdge+self.itemPadding, 0),  # y pos irrelevant
                size=[w, 0.1],  # expand height with text
                autoLog=False,
                colorSpace=self.colorSpace,
                color=item['itemColor'] or self.itemColor,
                fillColor=None,
                padding=0,  # handle this by padding between items
                borderWidth=1,
                borderColor='red' if debug else None,  # add borderColor to help debug
                editable=False,
                bold=bold,
                font=item['font'] or self.font)
        # Resize textbox to be at least as tall as the text
        question._updateVertices()
        textHeight = getattr(question.boundingBox._size, question.units)[1]
        if textHeight > question.size[1]:
            question.size[1] = textHeight + question.padding[1] * 2
            question._layout()

        questionHeight = question.size[1]
        questionWidth = question.size[0]
        # store virtual pos to combine with scroll bar for actual pos
        question._baseY = self._currentVirtualY

        # Add question objects to Form element dict
        item['itemCtrl'] = question

        return question, questionHeight, questionWidth

    def _setResponse(self, item):
        """Makes calls to methods which make Slider or TextBox response objects
        for Form

        Parameters
        ----------
        item : dict
            The dict entry for a single item
        question : TextStim
            The question text object

        Returns
        -------
        psychopy.visual.slider.Slider
            The Slider object for response
        psychopy.visual.TextBox
            The TextBox object for response
        respHeight
            The height of the response object as type float
        """
        if self.autoLog:
            logging.info(
                    "Adding response to Form type: {}, layout: {}, options: {}"
                    .format(item['type'], item['layout'], item['options']))

        if item['type'].lower() == 'free text':
            respCtrl, respHeight = self._makeTextBox(item)
        elif item['type'].lower() in ['heading', 'description']:
            respCtrl, respHeight = None, 0
        elif item['type'].lower() in ['rating', 'slider', 'choice', 'radio']:
            respCtrl, respHeight = self._makeSlider(item)

        item['responseCtrl'] = respCtrl
        return respCtrl, float(respHeight)

    def _makeSlider(self, item):
        """Creates Slider object for Form class

        Parameters
        ----------
        item : dict
            The dict entry for a single item
        pos : tuple
            position of response object

        Returns
        -------
        psychopy.visual.slider.Slider
            The Slider object for response
        respHeight
            The height of the response object as type float
        """
        # Slider dict

        kind = item['type'].lower()

        # what are the ticks for the scale/slider?
        if item['type'].lower() in ['radio', 'choice']:
            if item['ticks']:
                ticks = item['ticks']
            else:
                ticks = None
            tickLabels = item['tickLabels'] or item['options'] or item['ticks']
            granularity = 1
            style = 'radio'
        else:
            if item['ticks']:
                ticks = item['ticks']
            elif item['options']:
                ticks = range(0, len(item['options']))

            else:
                raise ValueError("We don't appear to have either options or "
                                 "ticks for item '{}' of {}."
                                 .format(item['itemText'], self.name))
            # how to label those ticks
            if item['tickLabels']:
                tickLabels = [str(i).strip() for i in item['tickLabels']]
            elif 'options' in item and item['options']:
                tickLabels = [str(i).strip() for i in item['options']]
            else:
                tickLabels = None
            # style/granularity
            if kind == 'slider' and 'granularity' in item:
                if item['granularity']:
                    granularity = item['granularity']
                else:
                    granularity = 0
            elif kind == 'slider' and 'granularity' not in item:
                granularity = 0
            else:
                granularity = 1
            style = kind

        # Make invisible guide rect to help with laying out slider
        w = (item['responseWidth'] - self.itemPadding * 2) * (self.size[0] - self.scrollbarWidth) * 0.8
        if item['layout'] == 'horiz':
            h = self.textHeight * 2 + 0.03
        elif item['layout'] == 'vert':
            h = self.textHeight * 1.1 * len(item['options'])
        x = self.rightEdge - self.itemPadding - self.scrollbarWidth - w * 0.1
        guide = Rect(
            self.win,
            size=(w, h),
            pos=(x, 0),
            anchor="top-right",
            lineColor="red",
            fillColor=None,
            units=self.units,
            autoLog=False
        )
        # Get slider pos and size
        if item['layout'] == 'horiz':
            x = guide.pos[0] - guide.size[0] / 2
            w = guide.size[0]
            h = 0.03
            wrap = None  # Slider defaults are fine for horizontal
        elif item['layout'] == 'vert':
            # for vertical take into account the nOptions
            x = guide.pos[0] - guide.size[0]
            w = 0.03
            h = guide.size[1]
            wrap = guide.size[0] / 2 - 0.03
            item['options'].reverse()

        # Create Slider
        resp = psychopy.visual.Slider(
                self.win,
                pos=(x, 0),  # NB y pos is irrelevant here - handled later
                size=(w, h),
                ticks=ticks,
                labels=tickLabels,
                units=self.units,
                labelHeight=self.textHeight,
                labelWrapWidth=wrap,
                granularity=granularity,
                flip=True,
                style=style,
                autoLog=False,
                font=item['font'] or self.font,
                color=item['responseColor'] or self.responseColor,
                fillColor=item['markerColor'] or self.markerColor,
                borderColor=item['responseColor'] or self.responseColor,
                colorSpace=self.colorSpace)
        resp.guide = guide

        # store virtual pos to combine with scroll bar for actual pos
        resp._baseY = self._currentVirtualY - guide.size[1] / 2 - self.itemPadding

        return resp, guide.size[1]

    def _getItemHeight(self, item, ctrl=None):
        """Returns the full height of the item to be inserted in the form"""
        if type(ctrl) == psychopy.visual.TextBox2:
            return ctrl.size[1]
        if type(ctrl) == psychopy.visual.Slider:
            # Set radio button layout
            if item['layout'] == 'horiz':
                return 0.03 + ctrl.labelHeight*3
            elif item['layout'] == 'vert':
                # for vertical take into account the nOptions
                return ctrl.labelHeight*len(item['options'])

    def _makeTextBox(self, item):
        """Creates TextBox object for Form class

        NOTE: The TextBox 2 in work in progress, and has not been added to Form class yet.
        Parameters
        ----------
        item : dict
            The dict entry for a single item
        pos : tuple
            position of response object

        Returns
        -------
        psychopy.visual.TextBox
            The TextBox object for response
        respHeight
            The height of the response object as type float
        """
        w = (item['responseWidth'] - self.itemPadding * 2) * (self.size[0] - self.scrollbarWidth)
        x = self.rightEdge - self.itemPadding - self.scrollbarWidth
        resp = psychopy.visual.TextBox2(
                self.win,
                text='',
                pos=(x, 0),  # y pos irrelevant now (handled by scrollbar)
                size=(w, 0.1),
                letterHeight=self.textHeight,
                units=self.units,
                anchor='top-right',
                color=item['responseColor'] or self.responseColor,
                colorSpace=self.colorSpace,
                font=item['font'] or self.font,
                editable=True,
                borderColor=item['responseColor'] or self.responseColor,
                borderWidth=2,
                fillColor=None,
                onTextCallback=self._layoutY,
        )
        if debug:
            resp.borderColor = "red"
        # Resize textbox to be at least as tall as the text
        resp._updateVertices()
        textHeight = getattr(resp.boundingBox._size, resp.units)[1]
        if textHeight > resp.size[1]:
            resp.size[1] = textHeight + resp.padding[1] * 2
            resp._layout()

        respHeight = resp.size[1]
        # store virtual pos to combine with scroll bar for actual pos
        resp._baseY = self._currentVirtualY

        return resp, respHeight

    def _setScrollBar(self):
        """Creates Slider object for scrollbar

        Returns
        -------
        psychopy.visual.slider.Slider
            The Slider object for scroll bar
        """
        scroll = psychopy.visual.Slider(win=self.win,
                                        size=(self.scrollbarWidth, self.size[1] / 1.2),  # Adjust size to account for scrollbar overflow
                                        ticks=[0, 1],
                                        style='scrollbar',
                                        borderColor=self.responseColor,
                                        fillColor=self.markerColor,
                                        pos=(self.rightEdge - self.scrollbarWidth / 2, self.pos[1]),
                                        autoLog=False)
        return scroll

    def _setBorder(self):
        """Creates border using Rect

        Returns
        -------
        psychopy.visual.Rect
            The border for the survey
        """
        return psychopy.visual.Rect(win=self.win,
                                    units=self.units,
                                    pos=self.pos,
                                    width=self.size[0],
                                    height=self.size[1],
                                    colorSpace=self.colorSpace,
                                    fillColor=self.fillColor,
                                    lineColor=self.borderColor,
                                    opacity=None,
                                    autoLog=False)

    def _setAperture(self):
        """Blocks text beyond border using Aperture

        Returns
        -------
        psychopy.visual.Aperture
            The aperture setting viewable area for forms
        """
        aperture = psychopy.visual.Aperture(win=self.win,
                                            name='aperture',
                                            units=self.units,
                                            shape='square',
                                            size=self.size,
                                            pos=self.pos,
                                            autoLog=False)
        aperture.disable()  # Disable on creation. Only enable on draw.
        return aperture

    def _getScrollOffset(self):
        """Calculate offset position of items in relation to markerPos. Offset is a proportion of
        `vheight - height`, meaning the max offset (when scrollbar.markerPos is 1) is enough
        to take the bottom element to the bottom of the border.

        Returns
        -------
        float
            Offset position of items proportionate to scroll bar
        """
        offset = max(self._vheight - self.size[1], 0) * (1 - self.scrollbar.markerPos) * -1
        return offset

    def _createItemCtrls(self):
        """Define layout of form"""
        # Define boundaries of form
        if self.autoLog:
            logging.info("Setting layout of Form: {}.".format(self.name))

        self.leftEdge = self.pos[0] - self.size[0] / 2.0
        self.rightEdge = self.pos[0] + self.size[0] / 2.0

        # For each question, create textstim and rating scale
        for item in self.items:
            # set up the question object
            self._setQuestion(item)
            # set up the response object
            self._setResponse(item)


        # position a slider on right-hand edge
        self.scrollbar = self._setScrollBar()
        self.scrollbar.markerPos = 1  # Set scrollbar to start position
        self.border = self._setBorder()
        self.aperture = self._setAperture()
        # then layout the Y positions
        self._layoutY()

        if self.autoLog:
            logging.info("Layout set for Form: {}.".format(self.name))

    def _layoutY(self):
        """This needs to be done when editable textboxes change their size
        because everything below them needs to move too"""

        self.topEdge = self.pos[1] + self.size[1] / 2.0

        self._currentVirtualY = self.topEdge - self.itemPadding
        # For each question, create textstim and rating scale
        for item in self.items:
            question = item['itemCtrl']
            response = item['responseCtrl']

            # update item baseY
            question._baseY = self._currentVirtualY
            # and get height to update current Y
            questionHeight = self._getItemHeight(item=item, ctrl=question)

            # go on to next line if together they're too wide
            oneLine = (item['itemWidth']+item['responseWidth'] <= 1
                       or not response)
            if not oneLine:
                # response on next line
                self._currentVirtualY -= questionHeight + self.itemPadding / 4

            # update response baseY
            if not response:
                self._currentVirtualY -= questionHeight + self.itemPadding
                continue
            # get height to update current Y
            respHeight = self._getItemHeight(item=item, ctrl=response)

            # update item baseY
            # slider needs to align by middle
            if type(response) == psychopy.visual.Slider:
                response._baseY = self._currentVirtualY - max(questionHeight, respHeight)/2
            else:  # hopefully we have an object that can anchor at top?
                response._baseY = self._currentVirtualY

            # go on to next line if together they're too wide
            if oneLine:
                # response on same line - work out which is bigger
                self._currentVirtualY -= (
                    max(questionHeight, respHeight) + self.itemPadding
                )
            else:
                # response on next line
                self._currentVirtualY -= respHeight + self.itemPadding * 5/4

        # Calculate virtual height as distance from top edge to bottom of last element
        self._vheight = abs(self.topEdge - self._currentVirtualY)

        self._setDecorations()  # choose whether show/hide scroolbar

    def _setDecorations(self):
        """Sets Form decorations i.e., Border and scrollbar"""
        # add scrollbar if it's needed
        self._decorations = [self.border]
        if self._vheight > self.size[1]:
            self._decorations.append(self.scrollbar)

    def _inRange(self, item):
        """Check whether item position falls within border area

        Parameters
        ----------
        item : TextStim, Slider object
            TextStim or Slider item from survey

        Returns
        -------
        bool
            Returns True if item position falls within border area
        """
        upperRange = self.size[1]
        lowerRange = -self.size[1]
        return (item.pos[1] < upperRange and item.pos[1] > lowerRange)

    def _drawDecorations(self):
        """Draw decorations on form."""
        [decoration.draw() for decoration in self._decorations]

    def _drawExternalDecorations(self):
        """Draw decorations outside the aperture"""
        [decoration.draw() for decoration in self._externalDecorations]

    def _drawCtrls(self):
        """Draw elements on form within border range.

        Parameters
        ----------
        items : List
            List of TextStim or Slider item from survey
        """
        for idx, item in enumerate(self.items):
            for element in [item['itemCtrl'], item['responseCtrl']]:
                if element is None:  # e.g. because this has no resp obj
                    continue

                element.pos = (element.pos[0],
                               element._baseY - self._getScrollOffset())
                if self._inRange(element):
                    element.draw()
                    if debug and hasattr(element, "guide"):
                        # If debugging, draw position guide too
                        element.guide.pos = (element.guide.pos[0], element._baseY - self._getScrollOffset() + element.guide.size[1] / 2)
                        element.guide.draw()

    def setAutoDraw(self, value, log=None):
        """Sets autoDraw for Form and any responseCtrl contained within
        """
        for i in self.items:
            if i['responseCtrl']:
                i['responseCtrl'].__dict__['autoDraw'] = value
                self.win.addEditable(i['responseCtrl'])
        BaseVisualStim.setAutoDraw(self, value, log)

    def draw(self):
        """Draw all form elements"""
        # Check mouse wheel
        self.scrollbar.markerPos += self.scrollbar.mouse.getWheelRel()[
                                        1] / self.scrollSpeed
        # draw the box and scrollbar
        self._drawExternalDecorations()
        # enable aperture
        self.aperture.enable()
        # draw the box and scrollbar
        self._drawDecorations()
        # Draw question and response objects
        self._drawCtrls()
        # disable aperture
        self.aperture.disable()

    def getData(self):
        """Extracts form questions, response ratings and response times from
        Form items

        Returns
        -------
        list
            A copy of the data as a list of dicts
        """
        nIncomplete = 0
        nIncompleteRequired = 0
        for thisItem in self.items:
            if 'responseCtrl' not in thisItem or not thisItem['responseCtrl']:
                continue  # maybe a heading or similar
            responseCtrl = thisItem['responseCtrl']
            # get response if available
            if hasattr(responseCtrl, 'getRating'):
                thisItem['response'] = responseCtrl.getRating()
            else:
                thisItem['response'] = responseCtrl.text
            if thisItem['response'] in [None, '']:
                # todo : handle required items here (e.g. ending with * ?)
                nIncomplete += 1
            # get RT if available
            if hasattr(responseCtrl, 'getRT'):
                thisItem['rt'] = responseCtrl.getRT()
            else:
                thisItem['rt'] = None
        self._complete = (nIncomplete == 0)
        return copy.copy(self.items)  # don't want users changing orig

    def reset(self):
        """
        Clear all responses and set all items to their initial values.
        """
        # Iterate through all items
        for item in self.items:
            # If item doesn't have a response ctrl, skip it
            if "responseCtrl" not in item:
                continue
            # If response ctrl is a slider, set its rating to None
            if isinstance(item['responseCtrl'], psychopy.visual.Slider):
                item['responseCtrl'].rating = None
            # If response ctrl is a textbox, set its text to blank
            elif isinstance(item['responseCtrl'], psychopy.visual.TextBox2):
                item['responseCtrl'].text = ""
        # Set scrollbar to top
        self.scrollbar.rating = 1

    def addDataToExp(self, exp, itemsAs='rows'):
        """Gets the current Form data and inserts into an
        :class:`~psychopy.experiment.ExperimentHandler` object either as rows
        or as columns

        Parameters
        ----------
        exp : :class:`~psychopy.experiment.ExperimentHandler`
        itemsAs: 'rows','cols' (or 'columns')

        Returns
        -------

        """
        data = self.getData()  # will be a copy of data (we can trash it)
        asCols = itemsAs.lower() in ['cols', 'columns']
        # iterate over items and fields within each item
        # iterate all items and all fields before calling nextEntry
        for ii, thisItem in enumerate(data):  # data is a list of dicts
            for fieldName in thisItem:
                if fieldName in _doNotSave:
                    continue
                if asCols:  # for columns format, we need index for item
                    columnName = "{}[{}].{}".format(self.name, ii, fieldName)
                else:
                    columnName = "{}.{}".format(self.name, fieldName)
                exp.addData(columnName, thisItem[fieldName])
                # finished field
            if not asCols:  # for rows format we add a newline each item
                exp.nextEntry()
            # finished item
        # finished form
        if asCols:  # for cols format we add a newline each item
            exp.nextEntry()

    def formComplete(self):
        """Deprecated in version 2020.2. Please use the Form.complete property
        """
        return self.complete

    @property
    def pos(self):
        if hasattr(self, '_pos'):
            return self._pos

    @pos.setter
    def pos(self, value):
        self._pos = value
        if hasattr(self, 'aperture'):
            self.aperture.pos = value
        if hasattr(self, 'border'):
            self.border.pos = value
        self.leftEdge = self.pos[0] - self.size[0] / 2.0
        self.rightEdge = self.pos[0] + self.size[0] / 2.0
        # Set horizontal position of elements
        for item in self.items:
            for element in [item['itemCtrl'], item['responseCtrl']]:
                if element is None:  # e.g. because this has no resp obj
                    continue
                element.pos = [value[0], element.pos[1]]
                element._baseY = value[1]
                if hasattr(element, 'anchor'):
                    element.anchor = 'top-center'
        # Calculate new position for everything on the y axis
        self.scrollbar.pos = (self.rightEdge - .008, self.pos[1])
        self._layoutY()

    @property
    def scrollbarWidth(self):
        """
        Width of the scrollbar for this Form, in the spatial units of this Form. Can also be set as a
        `layout.Vector` object.
        """
        if not hasattr(self, "_scrollbarWidth"):
            # Default to 15px
            self._scrollbarWidth = layout.Vector(15, 'pix', self.win)
        return getattr(self._scrollbarWidth, self.units)[0]

    @scrollbarWidth.setter
    def scrollbarWidth(self, value):
        self._scrollbarWidth = layout.Vector(value, self.units, self.win)
        self.scrollbar.width[0] = self.scrollbarWidth

    @property
    def opacity(self):
        return BaseVisualStim.opacity.fget(self)

    @opacity.setter
    def opacity(self, value):
        BaseVisualStim.opacity.fset(self, value)
        self.fillColor = self._fillColor
        self.borderColor = self._borderColor
        if hasattr(self, "_foreColor"):
            self._foreColor.alpha = value
        if hasattr(self, "_itemColor"):
            self._itemColor.alpha = value
        if hasattr(self, "_responseColor"):
            self._responseColor.alpha = value
        if hasattr(self, "_markerColor"):
            self._markerColor.alpha = value

    @property
    def complete(self):
        """A read-only property to determine if the current form is complete"""
        self.getData()
        return self._complete

    @property
    def foreColor(self):
        """
        Sets both `itemColor` and `responseColor` to the same value
        """
        return ColorMixin.foreColor.fget(self)

    @foreColor.setter
    def foreColor(self, value):
        ColorMixin.foreColor.fset(self, value)
        self.itemColor = value
        self.responseColor = value

    @property
    def fillColor(self):
        """
        Color of the form's background
        """
        return ColorMixin.fillColor.fget(self)

    @fillColor.setter
    def fillColor(self, value):
        ColorMixin.fillColor.fset(self, value)
        if hasattr(self, "border"):
            self.border.fillColor = value

    @property
    def borderColor(self):
        """
        Color of the line around the form
        """
        return ColorMixin.borderColor.fget(self)

    @borderColor.setter
    def borderColor(self, value):
        ColorMixin.borderColor.fset(self, value)
        if hasattr(self, "border"):
            self.border.borderColor = value

    @property
    def itemColor(self):
        """
        Color of the text on form items
        """
        return getattr(self._itemColor, self.colorSpace)

    @itemColor.setter
    def itemColor(self, value):
        self._itemColor = Color(value, self.colorSpace)
        # Set text color on each item
        for item in self.items:
            if 'itemCtrl' in item:
                if isinstance(item['itemCtrl'], psychopy.visual.TextBox2):
                    item['itemCtrl'].foreColor =  self._itemColor

    @property
    def responseColor(self):
        """
        Color of the lines and text on form responses
        """
        if hasattr(self, "_responseColor"):
            return getattr(self._responseColor, self.colorSpace)

    @responseColor.setter
    def responseColor(self, value):
        self._responseColor = Color(value, self.colorSpace)
        # Set line color on scrollbar
        if hasattr(self, "scrollbar"):
            self.scrollbar.borderColor = self._responseColor
        # Set line and label color on each item
        for item in self.items:
            if 'responseCtrl' in item:
                if isinstance(item['responseCtrl'], psychopy.visual.Slider) or isinstance(item['responseCtrl'], psychopy.visual.TextBox2):
                    item['responseCtrl'].borderColor = self._responseColor
                    item['responseCtrl'].foreColor = self._responseColor

    @property
    def markerColor(self):
        """
        Color of the marker on any sliders in this form
        """
        if hasattr(self, "_markerColor"):
            return getattr(self._markerColor, self.colorSpace)

    @markerColor.setter
    def markerColor(self, value):
        self._markerColor = Color(value, self.colorSpace)
        # Set marker color on scrollbar
        if hasattr(self, "scrollbar"):
            self.scrollbar.fillColor = self._markerColor
        # Set marker color on each item
        for item in self.items:
            if 'responseCtrl' in item:
                if isinstance(item['responseCtrl'], psychopy.visual.Slider):
                    item['responseCtrl'].fillColor = self._markerColor

    @property
    def style(self):
        if hasattr(self, "_style"):
            return self._style

    @style.setter
    def style(self, style):
        """Sets some predefined styles or use these to create your own.

        If you fancy creating and including your own styles that would be great!

        Parameters
        ----------
        style: string

            Known styles currently include:

                'light': black text on a light background
                'dark': white text on a dark background

        """
        self._style = style
        # If style is custom, skip the rest
        if style in ['custom...', 'None', None]:
            return
        # If style is a string of a known style, use that
        if style in self.knownStyles:
            style = self.knownStyles[style]
        # By here, style should be a dict
        if not isinstance(style, dict):
            return
        # Apply each key in the style dict as an attr
        for key, val in style.items():
            if hasattr(self, key):
                setattr(self, key, val)

    @property
    def values(self):
        # Iterate through each control and append its value to a dict
        out = {}
        for item in self.getData():
            out.update(
                {item['index']: item['response']}
            )
        return out

    @values.setter
    def values(self, values):
        for item in self.items:
            if item['index'] in values:
                ctrl = item['responseCtrl']
                # set response if available
                if hasattr(ctrl, "rating"):
                    ctrl.rating = values[item['index']]
                elif hasattr(ctrl, "value"):
                    ctrl.value = values[item['index']]
                else:
                    ctrl.text = values[item['index']]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Creates a rectangle of given width and height as a special case of a
:class:`~psychopy.visual.ShapeStim`"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import numpy as np

import psychopy  # so we can get the __path__
from psychopy.visual.shape import BaseShapeStim
from psychopy.tools.attributetools import attributeSetter, setAttribute


class Rect(BaseShapeStim):
    """Creates a rectangle of given width and height as a special case of a
    :class:`~psychopy.visual.ShapeStim`.

    Parameters
    ----------
    win : :class:`~psychopy.visual.Window`
        Window this shape is being drawn to. The stimulus instance will
        allocate its required resources using that Windows context. In many
        cases, a stimulus instance cannot be drawn on different windows
        unless those windows share the same OpenGL context, which permits
        resources to be shared between them.
    width, height : float or int
        The width or height of the shape. *DEPRECATED* use `size` to define
        the dimensions of the shape on initialization. If `size` is
        specified the values of `width` and `height` are ignored. This is to
        provide legacy compatibility for existing applications.
    units : str
        Units to use when drawing. This will affect how parameters and
        attributes `pos`, `size` and `radius` are interpreted.
    lineWidth : float
        Width of the shape's outline.
    lineColor, fillColor : array_like, str, :class:`~psychopy.colors.Color` or None
        Color of the shape outline and fill. If `None`, a fully transparent
        color is used which makes the fill or outline invisible.
    lineColorSpace, fillColorSpace : str
        Colorspace to use for the outline and fill. These change how the
        values passed to `lineColor` and `fillColor` are interpreted.
        *Deprecated*. Please use `colorSpace` to set both outline and fill
        colorspace. These arguments may be removed in a future version.
    pos : array_like
        Initial position (`x`, `y`) of the shape on-screen relative to
        the origin located at the center of the window or buffer in `units`.
        This can be updated after initialization by setting the `pos`
        property. The default value is `(0.0, 0.0)` which results in no
        translation.
    size : array_like, float, int or None
        Width and height of the shape as `(w, h)` or `[w, h]`. If a single
        value is provided, the width and height will be set to the same
        specified value. If `None` is specified, the `size` will be set
        with values passed to `width` and `height`.
    ori : float
        Initial orientation of the shape in degrees about its origin.
        Positive values will rotate the shape clockwise, while negative
        values will rotate counterclockwise. The default value for `ori` is
        0.0 degrees.
    opacity : float
        Opacity of the shape. A value of 1.0 indicates fully opaque and 0.0
        is fully transparent (therefore invisible). Values between 1.0 and
        0.0 will result in colors being blended with objects in the
        background. This value affects the fill (`fillColor`) and outline
        (`lineColor`) colors of the shape.
    contrast : float
        Contrast level of the shape (0.0 to 1.0). This value is used to
        modulate the contrast of colors passed to `lineColor` and
        `fillColor`.
    depth : int
        Depth layer to draw the shape when `autoDraw` is enabled.
        *DEPRECATED*
    interpolate : bool
        Enable smoothing (anti-aliasing) when drawing shape outlines. This
        produces a smoother (less-pixelated) outline of the shape.
    lineRGB, fillRGB: array_like, :class:`~psychopy.colors.Color` or None
        *Deprecated*. Please use `lineColor` and `fillColor`. These
        arguments may be removed in a future version.
    name : str
        Optional name of the stimuli for logging.
    autoLog : bool
        Enable auto-logging of events associated with this stimuli. Useful
        for debugging and to track timing when used in conjunction with
        `autoDraw`.
    autoDraw : bool
        Enable auto drawing. When `True`, the stimulus will be drawn every
        frame without the need to explicitly call the
        :py:meth:`~psychopy.visual.shape.ShapeStim.draw()` method.
    color : array_like, str, :class:`~psychopy.colors.Color` or None
        Sets both the initial `lineColor` and `fillColor` of the shape.
    colorSpace : str
        Sets the colorspace, changing how values passed to `lineColor` and
        `fillColor` are interpreted.

    Attributes
    ----------
    width, height : float or int
        The width and height of the rectangle. Values are aliased with fields
        in the `size` attribute. Use these values to adjust the size of the
        rectangle in a single dimension after initialization.

    """
    def __init__(self,
                 win,
                 width=.5,
                 height=.5,
                 units='',
                 lineWidth=1.5,
                 lineColor=None,
                 lineColorSpace=None,
                 fillColor='white',
                 fillColorSpace=None,
                 pos=(0, 0),
                 size=None,
                 anchor=None,
                 ori=0.0,
                 opacity=None,
                 contrast=1.0,
                 depth=0,
                 interpolate=True,
                 lineRGB=False,
                 fillRGB=False,
                 name=None,
                 autoLog=None,
                 autoDraw=False,
                 color=None,
                 colorSpace='rgb'):
        # width and height attributes, these are later aliased with `size`
        self.__dict__['width'] = float(width)
        self.__dict__['height'] = float(height)

        # If the size argument was specified, override values of width and
        # height, this is to maintain legacy compatibility. Args width and
        # height should be deprecated in later releases.
        if size is None:
            size = (self.__dict__['width'],
                    self.__dict__['height'])

        # vertices for rectangle, CCW winding order
        vertices = np.array([[-.5,  .5],
                             [ .5,  .5],
                             [ .5, -.5],
                             [-.5, -.5]])
        super(Rect, self).__init__(
            win,
            units=units,
            lineWidth=lineWidth,
            lineColor=lineColor,
            lineColorSpace=lineColorSpace,
            fillColor=fillColor,
            fillColorSpace=fillColorSpace,
            vertices=vertices,
            closeShape=True,
            pos=pos,
            size=size,
            anchor=anchor,
            ori=ori,
            opacity=opacity,
            contrast=contrast,
            depth=depth,
            interpolate=interpolate,
            lineRGB=lineRGB,
            fillRGB=fillRGB,
            name=name,
            autoLog=autoLog,
            autoDraw=autoDraw,
            color=color,
            colorSpace=colorSpace)

    def setSize(self, size, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message

        :ref:`Operations <attrib-operations>` supported.
        """
        setAttribute(self, 'size', size, log, operation)

    def setWidth(self, width, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'width', width, log, operation)

    def setHeight(self, height, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'height', height, log, operation)



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 23 11:18:47 2013

Tests the psychopy.core.getTime Function:
  On Windows:
    1) Checks that the currently used core.getTime() implementation gives duration results
       consistent with directly using the Python high resolution timer.
    2) Checks that the time between core.getTime calls is never negative.
    3) Tests the overhead in the core.getTime function call
    4) Tries to assess the resolution of the underlying core.getTime() time base and the Python high resolution timer.
    5) Tests MonotonicClock, Clock, CountdownTimer, wait
@author: Sol

-----
Jan 2014, Jeremy Gray:
- Coverage of .quit, .shellCall, and increased coverage of StaticPeriod()
"""

import time
import sys
import numpy as np
import gc
import pytest

import psychopy
import psychopy.logging as logging
from psychopy.visual import Window
from psychopy.core import (getTime, MonotonicClock, Clock, CountdownTimer, wait,
                           StaticPeriod, shellCall)
from psychopy.clock import monotonicClock
from psychopy.tests import _vmTesting


def test_EmptyFunction():
    pass

PRINT_TEST_RESULTS = False


def printf(*args):
    if PRINT_TEST_RESULTS:
        for a in args:
            sys.stdout(a)
        print('')


py_time = None
py_timer_name = None

py_time=time.time
py_timer_name = 'time.time'


def printExceptionDetails():
    import traceback,pprint
    exc_type, exc_value, exc_traceback = sys.exc_info()
    pprint.pprint(exc_type, indent=1, width=80, depth=None)
    pprint.pprint(exc_value, indent=1, width=80, depth=None)
    pprint.pprint(traceback.format_tb(exc_traceback), indent=1, width=80, depth=None)


@pytest.mark.slow
def test_DelayDurationAccuracy(sample_size=100):
    # test with sample_size randomly selected durations between 0.05 and 1.0 msec
    durations=np.zeros((3,sample_size))
    durations[0,:] = (np.random.randint(50, 1001, sample_size) * 0.001)

    for t in range(sample_size):
        cdur=durations[0][t]
        start_times=py_time(),getTime()
        stime=start_times[0]
        while py_time()-stime<cdur-0.02:
            end_times=py_time(),getTime()
        while py_time()-stime<cdur:
            end_times=py_time(),getTime()
        durations[1][t]=end_times[0]-start_times[0]
        durations[2][t]=end_times[1]-start_times[1]

    clockDurVsExpected=durations[1]-durations[0]
    clockDurVsQpc=durations[1]-durations[2]

    printf("## %s vs. psychopy getTime() Duration Difference Test:\n"%(py_timer_name))
    printf(">> Actual Vs. Expected %s Duration Diffs (msec.usec):"%(py_timer_name))
    printf("\tmin:\t\t%.3f"%(clockDurVsExpected.min()*1000.0))
    printf("\tmax:\t\t%.3f"%(clockDurVsExpected.max()*1000.0))
    printf("\tmean:\t\t%.3f"%(clockDurVsExpected.mean()*1000.0))
    printf("\tstd:\t\t%.3f"%(clockDurVsExpected.std()*1000.0))
    printf(">> %s vs getTime Duration Diffs (msec.usec):"%(py_timer_name))
    printf("\tmin:\t\t%.3f"%(clockDurVsQpc.min()*1000.0))
    printf("\tmax:\t\t%.3f"%(clockDurVsQpc.max()*1000.0))
    printf("\tmean:\t\t%.3f"%(clockDurVsQpc.mean()*1000.0))
    printf("\tstd:\t\t%.3f"%(clockDurVsQpc.std()*1000.0))

    # check that the differences between time.clock and psychopy.getTime
    # (which is using Win QPC) are within these limits:
    #
    # fabs(min) or max diff:    < 50 usec
    # mean diff:                < 10 usec
    # std of diff:              < 5 usec
    try:
        assert np.fabs(clockDurVsQpc.min())<0.00005
        assert clockDurVsQpc.max()<0.00005
        assert np.fabs(clockDurVsQpc.mean())<0.00001
        assert clockDurVsQpc.std()<0.000005
        printf("\nDuration Difference Test: PASSED")
    except Exception:
        printf("\nDuration Difference Test: FAILED")
    printf("-------------------------------------\n")


@pytest.mark.slow
def test_TimebaseQuality(sample_size=1000):
    gc.disable()

    callTimes=np.zeros((5,sample_size))

    timer_clock_jumpbacks=0
    core_getTime_jumpbacks=0

    for t in range(sample_size):
       s=py_time()
       e=py_time()
       callTimes[0][t]=e-s
       if e<s:
           timer_clock_jumpbacks+=1

       s=getTime()
       e=getTime()
       callTimes[1][t]=e-s
       if e<s:
           core_getTime_jumpbacks+=1

       s=py_time()
       x=test_EmptyFunction()
       e=py_time()
       callTimes[2][t]=e-s

       s=py_time()
       x=py_time()
       e=py_time()
       callTimes[3][t]=e-s

       s=py_time()
       x=getTime()
       e=py_time()
       callTimes[4][t]=e-s


    gc.enable()

    printf("## Timebase 'Quality' Tests :\n")
    test_headers=(">> %s Resolution (msec.usec):"%(py_timer_name),
                  ">> core.getTime() Resolution (msec.usec):",
                  ">> Empty function (msec.usec):",
                  ">> %s (msec.usec):"%(py_timer_name),
                  ">> core.getTime() (msec.usec):")
    for i,header in enumerate(test_headers):
        printf(header)
        printf("\tmin:\t\t%.9f"%(callTimes[i].min()*1000.0))
        printf("\tmax:\t\t%.6f"%(callTimes[i].max()*1000.0))
        printf("\tmean:\t\t%.6f"%(callTimes[i].mean()*1000.0))
        printf("\tstd:\t\t%.6f"%(callTimes[i].std()*1000.0))

    printf(">> %s jumpbacks: "%(py_timer_name),timer_clock_jumpbacks)
    printf(">> core.getTime() jumpbacks: ",core_getTime_jumpbacks)

    # Test that these conditions are true:
    #   - Effective Resolution (mean inter timer call duration) of timer is < 10 usec
    #   - Maximum inter timer call duration is < 100 usec
    #   - no negative durations in timer call durations
    try:
        assert (callTimes[0].mean()*1000.0)<0.01
        assert (callTimes[0].max()*1000.0)<0.1
        assert timer_clock_jumpbacks==0
        printf("\n%s Call Time / Resolution Test: PASSED"%(py_timer_name))
    except Exception:
        printf("\n%s Call Time / Resolution Test: FAILED"%(py_timer_name))

    try:
        assert (callTimes[1].mean()*1000.0)<0.01
        assert (callTimes[1].max()*1000.0)<0.1
        assert core_getTime_jumpbacks==0
        printf("\ncore.getTime() Call Time / Resolution Test: PASSED")
    except Exception:
        printf("\ncore.getTime() Call Time / Resolution Test: FAILED")

    printf("-------------------------------------\n")


def test_MonotonicClock():
    try:
        mc = MonotonicClock()

        t1=mc.getTime()
        time.sleep(1.0)
        t2=mc.getTime()

        startTime=mc.getLastResetTime()

        assert t2>t1
        assert t2-t1 > 0.95
        assert t2-t1 < 1.05
        assert startTime > 0

        # Test things that 'should fail':
        try:
            x=mc.timeAtLastReset
            assert 1=="MonotonicClock should not have an attribute called 'timeAtLastReset'."
        except Exception:
            pass

        try:
            x=mc.reset()
            assert 1=="MonotonicClock should not have a method 'reset()'."
        except Exception:
            pass

        try:
            x=mc.add()
            assert 1=="MonotonicClock should not have a method 'add()'."
        except Exception:
            pass

        printf(">> MonotonicClock Test: PASSED")

    except Exception:
        printf(">> MonotonicClock Test: FAILED")
        printExceptionDetails()

    printf("-------------------------------------\n")


def test_Clock():
    try:
        c = Clock()

        t1=c.getTime()
        time.sleep(1.0)
        t2=c.getTime()

        startTime=c.getLastResetTime()

        assert t2>t1
        assert t2-t1 > 0.95
        assert t2-t1 < 1.05
        assert startTime > 0

        c.reset()
        t=c.getTime()
        assert t < 0.01

        c.reset(10)
        t=c.getTime()
        assert t > -10.0
        assert t < -9.9

        t1=c.getTime()
        c.add(50)
        t2=c.getTime()
        assert t2-t1 > -50.0
        assert t2-t1 < -49.9

        printf(">> Clock Test: PASSED")

    except Exception:
        printf(">> Clock Test: FAILED")
        printExceptionDetails()

    printf("-------------------------------------\n")


def test_CountdownTimer():
    try:
        cdt = CountdownTimer(5.0)

        assert cdt.getTime() <= 5.0
        assert cdt.getTime() >= 4.75

        time.sleep(cdt.getTime())

        assert np.fabs(cdt.getTime()) < 0.1

        printf(">> CountdownTimer Test: PASSED")

    except Exception:
        printf(">> CountdownTimer Test: FAILED")
        printExceptionDetails()

    printf("-------------------------------------\n")


def test_Wait(duration=1.55):
    try:
        t1=getTime()
        wait(duration)
        t2=getTime()

        # Check that the actual duration of the wait was close to the requested delay.
        #
        # Note that I have had to set this to a relatively high value of
        # 50 msec because on my Win7, i7, 16GB machine I would get delta's of up to
        # 35 msec when I was testing this.
        #
        # This is 'way high', and I think is because the current wait()
        # implementation polls pyglet for events during the CPUhog period.
        # IMO, during the hog period, which should only need to be only 1 - 2 msec
        # , not the 200 msec default now, nothing should be done but tight looping
        # waiting for the wait() to expire. This is what I do in ioHub and on this same
        # PC I get actual vs. requested duration delta's of < 100 usec consistently.
        #
        # I have not changed the wait in psychopy until feedback is given, as I
        # may be missing a reason why the current wait() implementation is required.
        #
        assert np.fabs((t2-t1)-duration) < 0.05

        printf(">> core.wait(%.2f) Test: PASSED"%(duration))

    except Exception:
        printf(">> core.wait(%.2f) Test: FAILED. Actual Duration was %.3f"%(duration,(t2-t1)))
        printExceptionDetails()

    printf("-------------------------------------\n")


def test_LoggingDefaultClock():
    try:
        t1 = logging.defaultClock.getTime()
        t2 = getTime()
        t3 = monotonicClock.getTime()

        assert np.fabs(t1-t2) < 0.02
        assert np.fabs(t1-t3) < 0.02
        assert np.fabs(t3-t2) < 0.02

        assert logging.defaultClock.getLastResetTime() == monotonicClock.getLastResetTime()

        printf(">> logging.defaultClock Test: PASSED")

    except Exception:
        printf(">> logging.defaultClock Test: FAILED. ")
        printExceptionDetails()

    printf("-------------------------------------\n")


@pytest.mark.staticperiod
def test_StaticPeriod():
    static = StaticPeriod()
    static.start(0.1)
    wait(0.05)
    assert static.complete()==1
    static.start(0.1)
    wait(0.11)
    assert static.complete()==0

    win = Window(autoLog=False)
    static = StaticPeriod(screenHz=60, win=win)
    static.start(.002)
    assert win.recordFrameIntervals is False
    static.complete()
    assert static._winWasRecordingIntervals == win.recordFrameIntervals
    win.close()

    # Test if screenHz parameter is respected, i.e., if after completion of the
    # StaticPeriod, 1/screenHz seconds are still remaining, so the period will
    # complete after the next flip.
    refresh_rate = 100.0
    period_duration = 0.1
    timer = CountdownTimer()
    win = Window(autoLog=False)

    static = StaticPeriod(screenHz=refresh_rate, win=win)
    static.start(period_duration)
    timer.reset(period_duration )
    static.complete()

    if _vmTesting:
        tolerance = 0.01  # without a proper screen timing might not eb sub-ms
    else:
        tolerance = 0.001
    assert np.allclose(timer.getTime(),
                       1.0/refresh_rate,
                       atol=tolerance)
    win.close()


@pytest.mark.quit
def test_quit():
    # to-do: make some active threads
    with pytest.raises(SystemExit):
        psychopy.core.quit()


@pytest.mark.shellCall
class Test_shellCall():
    def setup_class(self):
        if sys.platform == 'win32':
            self.cmd = 'findstr'
        else:
            self.cmd = 'grep'

        self.msg = 'echo'

    def test_invalid_argument(self):
        with pytest.raises(TypeError):
            shellCall(12345)

    def test_stdin(self):
        echo = shellCall([self.cmd, self.msg], stdin=self.msg)
        assert echo == self.msg

        echo = shellCall(self.cmd + ' ' + self.msg, stdin=self.msg)
        assert echo == self.msg

    def test_stderr(self):
        _, se = shellCall([self.cmd, self.msg], stderr=True)
        assert se == ''

        _, se = shellCall(self.cmd + ' ' + self.msg, stderr=True)
        assert se == ''

    def test_stdin_and_stderr(self):
        echo, se = shellCall([self.cmd, self.msg], stdin='echo', stderr=True)
        assert echo == self.msg
        assert se == ''

        echo, se = shellCall(self.cmd + ' ' + self.msg, stdin='echo',
                             stderr=True)
        assert echo == self.msg
        assert se == ''

    def test_encoding(self):
        shellCall([self.cmd, self.msg], stdin=self.msg, encoding='utf-8')


if __name__ == '__main__':
    test_MonotonicClock()
    test_Clock()
    test_CountdownTimer()
    test_Wait()
    test_LoggingDefaultClock()
    test_TimebaseQuality()
    test_StaticPeriod()
    printf("\n** Next Test will Take ~ 1 minute...**\n")
    test_DelayDurationAccuracy()



import pytest
from psychopy import visual
from .test_basevisual import _TestColorMixin, _TestUnitsMixin


class TestShape(_TestColorMixin, _TestUnitsMixin):

    @classmethod
    def setup_class(self):

        self.win = visual.Window([128,128], pos=[50,50], allowGUI=False, autoLog=False)
        self.obj = visual.Rect(self.win, units="pix", pos=(0, 0), size=(128, 128), lineWidth=10)

        # Pixel which is the border color
        self.borderPoint = (0, 0)
        self.borderUsed = True
        # Pixel which is the fill color
        self.fillPoint = (50, 50)
        self.fillUsed = True
        # Shape has no foreground color
        self.foreUsed = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simple iohub eye tracker device demo.
Select which tracker to use by setting the TRACKER variable below.
"""

from psychopy import core, visual
from psychopy.iohub import launchHubServer
from psychopy.iohub.util import hideWindow, showWindow

# Eye tracker to use ('mouse', 'eyelink', 'gazepoint', or 'tobii')
TRACKER = 'mouse'
BACKGROUND_COLOR = [128, 128, 128]

devices_config = dict()
eyetracker_config = dict(name='tracker')
if TRACKER == 'mouse':
    eyetracker_config['calibration'] = dict(screen_background_color=BACKGROUND_COLOR)
    devices_config['eyetracker.hw.mouse.EyeTracker'] = eyetracker_config
elif TRACKER == 'eyelink':
    eyetracker_config['model_name'] = 'EYELINK 1000 DESKTOP'
    eyetracker_config['runtime_settings'] = dict(sampling_rate=1000, track_eyes='RIGHT')
    eyetracker_config['calibration'] = dict(screen_background_color=BACKGROUND_COLOR)
    devices_config['eyetracker.hw.sr_research.eyelink.EyeTracker'] = eyetracker_config
elif TRACKER == 'gazepoint':
    eyetracker_config['calibration'] = dict(use_builtin=False, screen_background_color=BACKGROUND_COLOR)
    devices_config['eyetracker.hw.gazepoint.gp3.EyeTracker'] = eyetracker_config
elif TRACKER == 'tobii':
    eyetracker_config['calibration'] = dict(screen_background_color=BACKGROUND_COLOR)
    devices_config['eyetracker.hw.tobii.EyeTracker'] = eyetracker_config
else:
    print("{} is not a valid TRACKER name; please use 'mouse', 'eyelink', 'gazepoint', or 'tobii'.".format(TRACKER))
    core.quit()

# Number if 'trials' to run in demo
TRIAL_COUNT = 2
# Maximum trial time / time timeout
T_MAX = 60.0
win = visual.Window((1920, 1080),
                    units='pix',
                    fullscr=True,
                    allowGUI=False,
                    colorSpace='rgb255',
                    monitor='55w_60dist',
                    color=BACKGROUND_COLOR
                    )

win.setMouseVisible(False)
text_stim = visual.TextStim(win, text="Start of Experiment",
                            pos=[0, 0], height=24,
                            color='black', units='pix', colorSpace='named',
                            wrapWidth=win.size[0] * .9)

text_stim.draw()
win.flip()

io = launchHubServer(window=win, **devices_config)

# Get some iohub devices for future access.
keyboard = io.getDevice('keyboard')
tracker = io.getDevice('tracker')

# Minimize the PsychoPy window if needed
hideWindow(win)
# Display calibration gfx window and run calibration.
result = tracker.runSetupProcedure()
print("Calibration returned: ", result)
# Maximize the PsychoPy window if needed
showWindow(win)

gaze_ok_region = visual.Circle(win, lineColor='black', radius=300, units='pix', colorSpace='named')

gaze_dot = visual.GratingStim(win, tex=None, mask='gauss', pos=(0, 0),
                              size=(40, 40), color='green', colorSpace='named', units='pix')

text_stim_str = 'Eye Position: %.2f, %.2f. In Region: %s\n'
text_stim_str += 'Press space key to start next trial.'
missing_gpos_str = 'Eye Position: MISSING. In Region: No\n'
missing_gpos_str += 'Press space key to start next trial.'
text_stim.setText(text_stim_str)

# Run Trials.....
t = 0
while t < TRIAL_COUNT:
    io.clearEvents()
    tracker.setRecordingState(True)
    run_trial = True
    tstart_time = core.getTime()
    while run_trial is True:
        # Get the latest gaze position in display coord space.
        gpos = tracker.getLastGazePosition()
        # Update stim based on gaze position
        valid_gaze_pos = isinstance(gpos, (tuple, list))
        gaze_in_region = valid_gaze_pos and gaze_ok_region.contains(gpos)
        if valid_gaze_pos:
            # If we have a gaze position from the tracker, update gc stim and text stim.
            if gaze_in_region:
                gaze_in_region = 'Yes'
            else:
                gaze_in_region = 'No'
            text_stim.text = text_stim_str % (gpos[0], gpos[1], gaze_in_region)

            gaze_dot.setPos(gpos)
        else:
            # Otherwise just update text stim
            text_stim.text = missing_gpos_str

        # Redraw stim
        gaze_ok_region.draw()
        text_stim.draw()
        if valid_gaze_pos:
            gaze_dot.draw()

        # Display updated stim on screen.
        flip_time = win.flip()

        # Check any new keyboard char events for a space key.
        # If one is found, set the trial end variable.
        #
        if keyboard.getPresses(keys=' '):
            run_trial = False
        elif core.getTime()-tstart_time > T_MAX:
            run_trial = False
    win.flip()
    # Current Trial is Done
    # Stop eye data recording
    tracker.setRecordingState(False)
    t += 1

# All Trials are done
# End experiment
win.close()
tracker.setConnectionState(False)
core.quit()



from psychopy.monitors import calibTools
import numpy

yy=[0,0.2,0.4,0.8,1.0]
minLum=2.0
maxLum=100.0
gamma=2.2

xxTest=numpy.array([ 0.,          0.48115651,  0.65935329,  0.90354543,  1.        ])
def test_GammaInverse_Eq1():
    xx= calibTools.gammaInvFun(yy, minLum, maxLum, gamma, b=0, eq=1)
    assert numpy.allclose(xx,xxTest,0.0001)
def test_GammaInverse_Eq4():
    xx= calibTools.gammaInvFun(yy, minLum, maxLum, gamma, b=0, eq=4)
    assert numpy.allclose(xx,xxTest,0.0001)
